"""Dissect plugin system.

See ``dissect/target/plugins/general/example.py`` for an example plugin.
"""

from __future__ import annotations

import fnmatch
import functools
import importlib
import importlib.util
import os
import sys
import traceback
from collections.abc import Callable
from dataclasses import dataclass, field
from itertools import chain, zip_longest
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypeAlias

from flow.record import Record, RecordDescriptor

import dissect.target.plugins.os.default as default
from dissect.target.exceptions import PluginError, PluginNotFoundError, UnsupportedPluginError
from dissect.target.helpers import cache
from dissect.target.helpers.fsutil import has_glob_magic
from dissect.target.helpers.logging import get_logger
from dissect.target.helpers.record import EmptyRecord
from dissect.target.helpers.utils import StrEnum

if TYPE_CHECKING:
    from collections.abc import Iterator

    from typing_extensions import Self

    from dissect.target.filesystem import Filesystem
    from dissect.target.helpers.record import ChildTargetRecord
    from dissect.target.target import Target

log = get_logger(__name__)

MODULE_PATH = "dissect.target.plugins"
"""The base module path to the in-tree plugins."""

OS_MODULE_PATH = "dissect.target.plugins.os"

OUTPUTS = (
    "default",
    "record",
    "yield",
    "none",
)
"""The different output types supported by ``@export``."""


class OperatingSystem(StrEnum):
    ANDROID = "android"
    BSD = "bsd"
    CITRIX = "citrix-netscaler"
    ESXI = "esxi"
    FORTIOS = "fortios"
    IOS = "ios"
    LINUX = "linux"
    MACOS = "macos"
    OSX = "osx"  # legacy
    PROXMOX = "proxmox"
    UNIX = "unix"
    VYOS = "vyos"
    WINDOWS = "windows"


@dataclass(frozen=True, eq=True, slots=True)
class PluginDescriptor:
    module: str
    qualname: str
    namespace: str
    path: str
    findable: bool
    functions: list[str]
    exports: list[str]

    @property
    def cls(self) -> type[Plugin]:
        return load(self)


@dataclass(frozen=True, eq=True, slots=True)
class FunctionDescriptor:
    name: str
    namespace: str
    path: str
    exported: bool
    internal: bool
    findable: bool
    alias: bool
    output: str
    method_name: str
    module: str
    qualname: str

    @property
    def cls(self) -> type[Plugin]:
        return load(self)

    @property
    def func(self) -> Callable[..., Any]:
        return getattr(self.cls, self.method_name)

    @property
    def record(self) -> RecordDescriptor | list[RecordDescriptor] | None:
        return getattr(self.func, "__record__", None)

    @property
    def args(self) -> list[tuple[list[str], dict[str, Any]]]:
        return getattr(self.func, "__args__", [])


@dataclass(frozen=True, eq=True, slots=True)
class FailureDescriptor:
    module: str
    stacktrace: list[str]


@dataclass(frozen=True, slots=True)
class PluginDescriptorLookup:
    # All regular plugins
    # {"<module_path>": PluginDescriptor}
    __regular__: dict[str, PluginDescriptor] = field(default_factory=dict)
    # All OS plugins
    # {"<module_path>": PluginDescriptor}
    __os__: dict[str, PluginDescriptor] = field(default_factory=dict)
    # All child plugins
    # {"<module_path>": PluginDescriptor}
    __child__: dict[str, PluginDescriptor] = field(default_factory=dict)


@dataclass(frozen=True, slots=True)
class FunctionDescriptorLookup:
    # All regular plugins
    # {"<function_name>": {"<module_path>": FunctionDescriptor}}
    __regular__: dict[str, dict[str, FunctionDescriptor]] = field(default_factory=dict)
    # All OS plugins
    # {"<function_name>": {"<module_path>": FunctionDescriptor}}
    __os__: dict[str, dict[str, FunctionDescriptor]] = field(default_factory=dict)
    # All child plugins
    # {"<function_name>": {"<module_path>": FunctionDescriptor}}
    __child__: dict[str, dict[str, FunctionDescriptor]] = field(default_factory=dict)


_OSTree: TypeAlias = dict[str, "_OSTree"]


@dataclass(frozen=True, slots=True)
class PluginRegistry:
    # Plugin descriptor lookup
    __plugins__: PluginDescriptorLookup = field(default_factory=PluginDescriptorLookup)
    # Function descriptor lookup
    __functions__: FunctionDescriptorLookup = field(default_factory=FunctionDescriptorLookup)
    # OS plugin tree
    # {"<module_part>": {"<module_part>": {}}}
    __ostree__: _OSTree = field(default_factory=dict)
    # Failures
    __failed__: list[FailureDescriptor] = field(default_factory=list)


PLUGINS: PluginRegistry = PluginRegistry()
"""The plugin registry.

Note: It's very important that all values in this dictionary are serializable.
The plugin registry can be stored in a file and loaded later. Plain Python syntax is used to store the registry.
An exception is made for :class:`FailureDescriptor`, :class:`FunctionDescriptor` and :class:`PluginDescriptor`.
"""

GENERATED = False


def export(*args, **kwargs) -> Callable[..., Any]:
    """Decorator to be used on Plugin functions that should be exported.

    Supported keyword arguments:
        property (bool): Whether this export should be regarded as a property.
            Properties are implicitly cached.

        cache (bool): Whether the result of this function should be cached.

        record (RecordDescriptor): The :class:`flow.record.RecordDescriptor` for the records that this function yields.
            If multiple record types are yielded, specificy each descriptor in a list or tuple.
            If the records are dynamically made, use :func:`dissect.target.helpers.record.DynamicDescriptor` instead.

        output (str): The output type of this function. Must be one of:
        - default: Single return value
        - record: Yields records. Implicit when record argument is given.
        - yield: Yields printable values.
        - none: No return value. Plugin is responsible for output formatting and should return ``None``.

    The ``export`` decorator adds some additional private attributes to an exported method or property:
    - ``__output__``: The output type to expect for this function, this is the same as ``output``.
    - ``__record__``: The type of record to expect, this value is the same as ``record``.
    - ``__exported__``: set to ``True`` to indicate the method or property is exported.

    Raises:
        ValueError: if there was an invalid output type.

    Returns:
        An exported function from a plugin.
    """

    def decorator(obj: Callable[..., Any]) -> Callable[..., Any] | property:
        # Properties are implicitly cached
        # Important! Currently it's crucial that this is *always* called
        # See the comment in Plugin.__init_subclass__ for more detail regarding Plugin.__call__
        obj = cache.wrap(obj, no_cache=not kwargs.get("cache", True), cls=kwargs.get("cls"))

        output = kwargs.get("output", "default")
        if output not in OUTPUTS:
            options = ", ".join(OUTPUTS)
            raise ValueError(f'Invalid output method "{output}", must be one of {options}')

        record = kwargs.get("record")
        if record is not None:
            output = "record"

            if not isinstance(record, (list, RecordDescriptor)) and not (
                isinstance(record, (list, tuple)) and all(isinstance(r, RecordDescriptor) for r in record)
            ):
                raise TypeError("record must be an instance of RecordDescriptor or a list of RecordDescriptor")

        obj.__output__ = output
        obj.__record__ = list(record) if isinstance(record, tuple) else record
        obj.__exported__ = True

        if kwargs.get("property", False):
            obj = property(obj)

        return obj

    if len(args) == 1:
        return decorator(args[0])
    return decorator


def internal(*args, **kwargs) -> Callable[..., Any]:
    """Decorator to be used on plugin functions that should be internal only.

    Making a plugin internal means that it's only callable from the Python API and not through ``target-query``.

    This decorator adds the ``__internal__`` private attribute to a method or property.
    The attribute is always set to ``True``, to tell :func:`register` that it is an internal
    method or property.
    """

    def decorator(obj: Callable[..., Any]) -> Callable[..., Any] | property:
        obj.__internal__ = True
        if kwargs.get("property", False):
            obj = property(obj)
        return obj

    if len(args) == 1:
        return decorator(args[0])
    return decorator


def arg(*args, **kwargs) -> Callable[..., Any]:
    """Decorator to be used on Plugin functions that accept additional command line arguments.

    Command line arguments can be added using the ``@arg`` decorator.
    Arguments to this decorator are directly forwarded to the ``ArgumentParser.add_argument`` function of ``argparse``.
    Resulting arguments are passed to the function using kwargs.
    The keyword argument name must match the argparse argument name.

    This decorator adds the ``__args__`` private attribute to a method or property.
    This attribute holds all the command line arguments that were added to the plugin function.
    """

    def decorator(obj: Callable[..., Any]) -> Callable[..., Any]:
        if not hasattr(obj, "__args__"):
            obj.__args__ = []

        obj.__args__.insert(0, (args, kwargs))

        return obj

    return decorator


def alias(*args, **kwargs: dict[str, Any]) -> Callable[..., Any]:
    """Decorator to be used on :class:`Plugin` functions to register an alias of that function."""

    if not kwargs.get("name") and not args:
        raise ValueError("Missing argument 'name'")

    def decorator(obj: Callable[..., Any]) -> Callable[..., Any]:
        if not hasattr(obj, "__aliases__"):
            obj.__aliases__ = []

        if name := (kwargs.get("name") or args[0]):
            obj.__aliases__.append(name)

        return obj

    return decorator


def clone_alias(cls: type, attr: Callable[..., Any], alias: str) -> None:
    """Clone the given attribute to an alias in the provided class."""

    # Clone the function object
    clone = type(attr)(attr.__code__, attr.__globals__, alias, attr.__defaults__, attr.__closure__)
    clone.__kwdefaults__ = attr.__kwdefaults__

    # Copy some attributes
    functools.update_wrapper(clone, attr)
    if wrapped := getattr(attr, "__wrapped__", None):
        # update_wrapper sets a new wrapper, we want the original
        clone.__wrapped__ = wrapped

    # Update module path so we can fool inspect.getmodule with subclassed Plugin classes
    clone.__module__ = cls.__module__

    # Update the names
    clone.__name__ = alias
    clone.__qualname__ = f"{cls.__name__}.{alias}"

    setattr(cls, alias, clone)


class Plugin:
    """Base class for plugins.

    Plugins can optionally be namespaced by specifying the ``__namespace__``
    class attribute. Namespacing results in your plugin needing to be prefixed
    with this namespace when being called. For example, if your plugin has
    specified ``test`` as namespace and a function called ``example``, you must
    call your plugin with ``test.example``.

    A ``Plugin`` class has the following private class attributes:

    - ``__namespace__``
    - ``__record_descriptors__``

    With the following two being assigned in :func:`register`:

    - ``__functions__``
    - ``__exports__``

    Additionally, the methods and attributes of :class:`Plugin` receive more private attributes
    by using decorators.

    The :func:`export` decorator adds the following private attributes

    - ``__exported__``
    - ``__output__``: Set with the :func:`export` decorator.
    - ``__record__``: Set with the :func:`export` decorator.

    The :func:`internal` decorator and :class:`InternalPlugin` set the ``__internal__`` attribute.
    Finally. :func:`args` decorator sets the ``__args__`` attribute.

    The :func:`alias` decorator populates the ``__aliases__`` private attribute of :class:`Plugin` methods.
    Resulting clones of the :class:`Plugin` are populated with the boolean ``__alias__`` attribute set to ``True``.

    Args:
        target: The :class:`~dissect.target.target.Target` object to load the plugin for.
    """

    __namespace__: str = None
    """Defines the plugin namespace."""
    __record_descriptors__: list[RecordDescriptor] = None
    """Defines a list of :class:`~flow.record.RecordDescriptor` of the exported plugin functions."""
    __register__: bool = True
    """Determines whether this plugin will be registered."""
    __findable__: bool = True
    """Determines whether this plugin will be revealed when using search patterns.

    Some (meta)-plugins are not very suitable for wild cards on CLI or
    plugin searches, because they will produce duplicate records or results.
    For instance a plugin that offers the same functions as subplugins will
    produce redundant results when used with a wild card
    (browser.* -> browser.history + browser.*.history).
    """
    __functions__: list[str]
    """Internal. A list of all method names decorated with ``@internal`` or ``@export``."""
    __exports__: list[str]
    """Internal. A list of all method names decorated with ``@export``."""

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        record_descriptors = _get_descriptors_on_nonprivate_methods(cls)
        cls.__record_descriptors__ = record_descriptors
        # This is a bit tricky currently
        # cls.__call__ is the *function* Plugin.__call__, not from the subclass
        # export() currently will _always_ return a new object because it always calls ``cache.wrap(obj)``
        # This allows this to work, otherwise the Plugin.__call__ would get all the plugin attributes set on it
        cls.__call__ = export(output="record", record=record_descriptors, cache=False, cls=cls)(cls.__call__)

        # Do not register the "base" subclasses defined in this file
        if cls.__module__ != Plugin.__module__:
            register(cls)

    def __init__(self, target: Target):
        self.target = target

    def __getattr__(self, name: str, /) -> Any:
        # "Magic" attribute access is only allowed on namespace plugins, to allow for indirect nesting of namespaces
        # E.g. `__namespace__ = "foo.bar"`, we should try to resolve `bar` on a plugin instance of `foo`
        if not self.__namespace__:
            raise AttributeError(name)

        # Attempt to lookup the attribute using the namespace
        try:
            _, func = self.target.get_function(f"{self.__class__.__namespace__}.{name}")
        except PluginNotFoundError:
            raise AttributeError(name)

        return func

    def is_compatible(self) -> bool:
        """Perform a compatibility check with the target."""
        try:
            self.check_compatible()
        except UnsupportedPluginError:
            return False

        return True

    def check_compatible(self) -> None:
        """Perform a compatibility check with the target.

        This function should return ``None`` if the plugin is compatible with the current target (``self.target``).
        For example, check if a certain file exists. Otherwise it should raise an
        :class:`UnsupportedPluginError`.

        Raises:
            UnsupportedPluginError: If the plugin could not be loaded.
        """
        raise NotImplementedError

    def __call__(self, *args, **kwargs) -> Iterator[Record | Any]:
        """Return the records of all exported methods.

        Raises:
            PluginError: If the subclass is not a namespace plugin.
        """
        if not self.__namespace__:
            raise PluginError(f"Plugin {self.__class__.__name__} is not a callable")

        for method_name in self.__exports__:
            if method_name == "__call__":
                continue

            method = getattr(self, method_name)
            if getattr(method, "__output__", None) not in ("record", "yield"):
                continue

            try:
                yield from method()
            except Exception as e:
                self.target.log.error("Error while executing `%s.%s`: %s", self.__namespace__, method_name, e)  # noqa: TRY400
                self.target.log.debug("", exc_info=e)

    def get_paths(self) -> Iterator[Path]:
        """Return all artifact paths."""
        if self.target.is_direct:
            yield from self._get_paths_direct()
        else:
            yield from self._get_paths()

    def get_all_paths(self) -> Iterator[Path]:
        """Return all artifact and auxiliary paths.

        The implementation of this function will
        probably change in the future, but the interface
        should stay the same.
        """
        yield from self.get_paths()
        yield from self._get_auxiliary_paths()

    def _get_paths_direct(self) -> Iterator[Path]:
        """Return all paths as given by the user."""
        for path in self.target._loader.paths:
            yield self.target.fs.path(str(path))

    def _get_paths(self) -> Iterator[Path]:
        """Return all artifact files of interest to the plugin.

        To be implemented by the plugin subclass.
        """
        raise NotImplementedError

    def _get_auxiliary_paths(self) -> Iterator[Path]:
        """Return all auxiliary files of interest to the plugin.

        To be implemented by the plugin subclass.
        """
        raise NotImplementedError


def register(plugincls: type[Plugin]) -> None:
    """Register a plugin, and put related data inside :attr:`PLUGINS`.

    This function uses the following private attributes that are set using decorators:
        - ``__exported__``: Set in :func:`export`.
        - ``__internal__``: Set in :func:`internal`.

    Additionally, ``register`` sets the following private attributes on the `plugincls`:
        - ``__functions__``: A list of all the methods and properties that are ``__internal__`` or ``__exported__``.
        - ``__exports__``: A list of all the methods or properties that were explicitly exported.

    If a plugincls ``__register__`` attribute is set to ``False``, the plugin will not be registered, but the
    plugin will still be processed for the private attributes mentioned above.

    Args:
        plugincls: A plugin class to register.

    Raises:
        ValueError: If ``plugincls`` is not a subclass of :class:`Plugin`.
    """
    if not issubclass(plugincls, Plugin):
        raise TypeError("Not a subclass of Plugin")

    # Register the plugin in the correct tree
    __plugins__ = PLUGINS.__plugins__
    __functions__ = PLUGINS.__functions__
    if issubclass(plugincls, OSPlugin):
        plugin_index = __plugins__.__os__
        function_index = __functions__.__os__
    elif issubclass(plugincls, ChildTargetPlugin):
        plugin_index = __plugins__.__child__
        function_index = __functions__.__child__
    else:
        plugin_index = __plugins__.__regular__
        function_index = __functions__.__regular__

    exports = []
    functions = []
    module_path = _module_path(plugincls)

    # This enables plugin directories, e.g.:
    # <plugin>/_plugin.py
    # <plugin>/helpers.py
    module_path = module_path.removesuffix("._plugin")

    module_key = f"{module_path}.{plugincls.__qualname__}"

    if not issubclass(plugincls, ChildTargetPlugin):
        # First pass to resolve aliases
        for attr in _get_nonprivate_attributes(plugincls):
            for alias in getattr(attr, "__aliases__", []):
                clone_alias(plugincls, attr, alias)

        for attr in _get_nonprivate_attributes(plugincls):
            if isinstance(attr, property):
                attr = attr.fget

            if getattr(attr, "__generated__", False) and plugincls != plugincls.__nsplugin__:
                continue

            exported = getattr(attr, "__exported__", False)
            internal = getattr(attr, "__internal__", False)

            if exported or internal:
                functions.append(attr.__name__)
                if exported:
                    exports.append(attr.__name__)

                if plugincls.__register__:
                    name = attr.__name__
                    if plugincls.__namespace__:
                        name = f"{plugincls.__namespace__}.{name}"

                    path = f"{module_path}.{attr.__name__}"

                    descriptor = FunctionDescriptor(
                        name=name,
                        namespace=plugincls.__namespace__,
                        path=path,
                        exported=exported,
                        internal=internal,
                        findable=plugincls.__findable__,
                        alias=name in getattr(attr, "__aliases__", []),
                        output=getattr(attr, "__output__", "none"),
                        method_name=attr.__name__,
                        module=plugincls.__module__,
                        qualname=plugincls.__qualname__,
                    )

                    # Register the functions in the lookup
                    function_index.setdefault(name, {})[module_key] = descriptor

    if plugincls.__namespace__:
        # Namespaces are also callable, so register the namespace itself as well
        # NamespacePlugin needs to register itself for every additional subclass, so allow overwrites here
        functions.append("__call__")
        if exports:
            exports.append("__call__")

        if plugincls.__register__:
            descriptor = FunctionDescriptor(
                name=plugincls.__namespace__,
                namespace=plugincls.__namespace__,
                path=module_path,
                exported=len(exports) != 0,
                internal=len(functions) != 0 and len(exports) == 0,
                findable=plugincls.__findable__,
                alias=False,
                output=getattr(plugincls.__call__, "__output__", "none"),
                method_name="__call__",
                module=plugincls.__module__,
                qualname=plugincls.__qualname__,
            )

            function_index.setdefault(plugincls.__namespace__, {})[module_key] = descriptor

    # Update the class with the plugin attributes
    plugincls.__functions__ = functions
    plugincls.__exports__ = exports

    if plugincls.__register__:
        plugin_index[module_key] = PluginDescriptor(
            module=plugincls.__module__,
            qualname=plugincls.__qualname__,
            namespace=plugincls.__namespace__,
            path=module_path,
            findable=plugincls.__findable__,
            functions=functions,
            exports=exports,
        )

        if issubclass(plugincls, OSPlugin):
            # Also store the OS plugins in a tree by module path
            # This is used to filter plugins based on the OSPlugin subclass
            # We don't store anything at the end of the tree, as we only use the tree to check if a plugin is compatible

            # Also slightly modify the module key to allow for more efficient filtering later
            # This is done by removing the last two parts of the module key, which are the file name and the class name
            module_parts = module_key.split(".")
            if module_parts[-2] != "_os":
                log.warning("OS plugin modules should be named as <os>/_os.py: %s", module_key)

            obj = PLUGINS.__ostree__
            for part in module_parts[:-2]:
                obj = obj.setdefault(part, {})

        log.trace("Plugin registered: %s", module_key)


def _get_plugins() -> PluginRegistry:
    """Load the plugin registry, or generate it if it doesn't exist yet."""
    global PLUGINS, GENERATED

    if not GENERATED:
        try:
            from dissect.target.plugins._pluginlist import PLUGINS
        except ImportError:
            PLUGINS = generate()

        GENERATED = True

    return PLUGINS


def _module_path(cls: type[Plugin] | str) -> str:
    """Returns the module path relative to ``dissect.target.plugins``."""
    if issubclass(cls, Plugin):
        module = getattr(cls, "__module__", "")
    elif isinstance(cls, str):
        module = cls
    else:
        raise TypeError(f"Invalid argument type: {cls}")

    return module.replace(MODULE_PATH, "").lstrip(".")


@functools.lru_cache(256)
def _os_match(osfilter: type[OSPlugin], module_path: str) -> bool:
    """Check if the a plugin is compatible with the given OS filter."""
    module_parts = module_path.split(".")

    # DefaultOSPlugin is "compatible" with everything, even other `os.*` modules (except `_os`)
    if issubclass(osfilter, default._os.DefaultOSPlugin) and (len(module_parts) < 2 or module_parts[-2] != "_os"):
        return True

    os_parts = _module_path(osfilter).split(".")[:-1]

    obj = _get_plugins().__ostree__
    for plugin_part, os_part in zip_longest(module_parts, os_parts):
        if plugin_part not in obj:
            break

        if plugin_part != os_part:
            return False

        obj = obj[plugin_part]

    return True


def plugins(osfilter: type[OSPlugin] | None = None, *, index: str = "__regular__") -> Iterator[PluginDescriptor]:
    """Walk the plugin registry and return plugin descriptors.

    If ``osfilter`` is specified, only plugins related to the provided OSPlugin, or plugins
    with no OS relation are returned. If ``osfilter`` is ``None``, all plugins will be returned.

    One exception to this is if the ``osfilter`` is a (sub-)class of ``DefaultOSPlugin``, then plugins
    are returned as if no ``osfilter`` was specified.

    The ``index`` parameter can be used to specify the index to return plugins from. By default,
    this is set to return regular plugins. Other possible values are ``__os__`` and ``__child__``.
    These return :class:`OSPlugin` and :class:`ChildTargetPlugin` respectively.

    Args:
        osfilter: The optional :class:`OSPlugin` to filter the returned plugins on.
        index: The plugin index to return plugins from. Defaults to regular plugins.

    Yields:
        Plugin descriptors in the plugin registry based on the given filter criteria.
    """

    plugin_index: dict[str, PluginDescriptor] = getattr(_get_plugins().__plugins__, index, {})

    # This is implemented as a list comprehension for performance reasons!
    yield from (
        descriptor
        for module_path, descriptor in plugin_index.items()
        if (index != "__os__" and (osfilter is None or _os_match(osfilter, module_path)))
        or (index == "__os__" and (osfilter is None or osfilter.__module__ == descriptor.module))
    )


def os_plugins() -> Iterator[PluginDescriptor]:
    """Retrieve all OS plugin descriptors."""
    yield from plugins(index="__os__")


def child_plugins() -> Iterator[PluginDescriptor]:
    """Retrieve all child plugin descriptors."""
    yield from plugins(index="__child__")


def functions(osfilter: type[OSPlugin] | None = None, *, index: str = "__regular__") -> Iterator[FunctionDescriptor]:
    """Retrieve all function descriptors.

    Args:
        osfilter: The optional :class:`OSPlugin` to filter the returned functions on.
        index: The plugin index to return functions from. Defaults to regular functions.

    Yields:
        Function descriptors in the plugin registry based on the given filter criteria.
    """

    function_index: dict[str, dict[str, FunctionDescriptor]] = getattr(_get_plugins().__functions__, index, {})

    # This is implemented as a list comprehension for performance reasons!
    yield from (
        descriptor
        for function_index in function_index.values()
        for module_path, descriptor in function_index.items()
        if osfilter is None or _os_match(osfilter, module_path)
    )


def lookup(
    function_name: str, osfilter: type[OSPlugin] | None = None, *, index: str = "__regular__"
) -> Iterator[FunctionDescriptor]:
    """Lookup a function descriptor by function name.

    Args:
        func_name: Function name to lookup.
        osfilter: The optional ``OSPlugin`` to filter results with for compatibility.
        index: The plugin index to return plugins from. Defaults to regular functions.

    Yields:
        Function descriptors that match the given function name and filter criteria.
    """

    function_index: dict[str, FunctionDescriptor] = getattr(_get_plugins().__functions__, index, {}).get(
        function_name, {}
    )

    # This is implemented as a list comprehension for performance reasons!
    entries: Iterator[FunctionDescriptor] = (
        value for key, value in function_index.items() if osfilter is None or _os_match(osfilter, key)
    )

    yield from sorted(entries, key=lambda desc: desc.module.count("."), reverse=True)


def load(desc: FunctionDescriptor | PluginDescriptor) -> type[Plugin]:
    """Helper function that loads a plugin from a given function or plugin descriptor.

    Args:
        desc: Function descriptor as returned by :func:`plugin.lookup` or plugin descriptor
              as returned by :func:`plugin.plugins`.

    Returns:
        The plugin class.

    Raises:
        PluginError: Raised when any other exception occurs while trying to load the plugin.
    """
    module = desc.module

    try:
        obj = importlib.import_module(module)
        for part in desc.qualname.split("."):
            obj = getattr(obj, part)
    except Exception as e:
        raise PluginError(f"An exception occurred while trying to load a plugin: {module}") from e
    else:
        return obj


def os_match(target: Target, descriptor: PluginDescriptor) -> bool:
    """Check if a plugin descriptor is compatible with the target OS.

    Args:
        target: The target to check compatibility with.
        descriptor: The plugin descriptor to check compatibility for.
    """
    return _os_match(target._os_plugin, f"{descriptor.module}.{descriptor.qualname}")


def failed() -> list[FailureDescriptor]:
    """Return all plugins that failed to load."""
    return _get_plugins().__failed__


@functools.cache
def _generate_long_paths(os_filter: type[OSPlugin]) -> dict[str, list[FunctionDescriptor]]:
    """Generate a dictionary of all long paths to their function descriptors for the given functions."""
    __regular__ = functions(os_filter, index="__regular__")
    __os__ = functions(os_filter, index="__os__")

    paths = {}
    for descriptor in chain(__regular__, __os__):
        # Namespace plugins are callable so exclude the explicit __call__ method
        if descriptor.method_name == "__call__":
            continue

        paths.setdefault(descriptor.path, []).append(descriptor)

    return paths


def find_functions(
    patterns: str,
    target: Target | None = None,
    compatibility: bool = False,
    ignore_load_errors: bool = False,
    show_hidden: bool = False,
) -> tuple[list[FunctionDescriptor], set[str]]:
    """Finds exported plugin functions that match the target and the patterns.

    Given a target, a comma separated list of patterns and an optional compatibility flag,
    this function finds matching plugins, optionally checking compatibility.

    Returns:
        A tuple containing a list of matching function descriptors and a set of invalid patterns.
    """
    found = []

    __functions__ = _get_plugins().__functions__

    regular_functions = __functions__.__regular__
    os_functions = __functions__.__os__

    os_filter = target._os_plugin if target is not None else None

    invalid_functions = set()

    for pattern in patterns.split(","):
        if not pattern:
            continue

        exact_match = pattern in regular_functions
        exact_os_match = pattern in os_functions

        if exact_match or exact_os_match:
            matches = list(_filter_exact_match(pattern, os_filter, "__os__" if exact_os_match else "__regular__"))
        else:
            # If we don't have an exact function match, do a slower treematch
            matches = list(_filter_tree_match(pattern, os_filter, show_hidden=show_hidden))

        if matches:
            found.extend(matches)
        else:
            invalid_functions.add(pattern)

    result = _filter_compatible(found, target, ignore_load_errors) if compatibility and target is not None else found
    return list(result), invalid_functions


def find_functions_by_record_field_type(
    field_types: str | list[str],
    target: Target | None = None,
    compatibility: bool = False,
    ignore_load_errors: bool = False,
) -> Iterator[FunctionDescriptor]:
    """Find functions that yield records with a specific field type.

    Args:
        field_types: The field type to search for.
        target: The target to check compatibility with.
        compatibility: Whether to check compatibility with the target.
        ignore_load_errors: Whether to ignore load errors.
    """

    def _search(search_types: list[str]) -> Iterator[FunctionDescriptor]:
        for descriptor in functions():
            try:
                if descriptor.output != "record" or (record_descriptors := descriptor.record) is None:
                    continue
            except Exception:
                if ignore_load_errors:
                    continue
                raise

            for rd in record_descriptors if isinstance(record_descriptors, list) else [record_descriptors]:
                if any(type_name in search_types for type_name, _ in rd._field_tuples):
                    yield descriptor
                    break

    search_types = field_types if isinstance(field_types, list) else [field_types]
    if compatibility and target is not None:
        yield from _filter_compatible(_search(search_types), target, ignore_load_errors)
    else:
        yield from _search(search_types)


def _filter_exact_match(
    pattern: str,
    os_filter: type[OSPlugin] | None,
    index: str = "__regular__",
) -> Iterator[FunctionDescriptor]:
    descriptors = lookup(pattern, os_filter, index=index)
    yield from (descriptor for descriptor in descriptors if descriptor.exported)


def _filter_tree_match(
    pattern: str, os_filter: type[OSPlugin] | None, show_hidden: bool = False
) -> Iterator[FunctionDescriptor]:
    """Perform a slow tree match on all ``__regular__`` and ``__os__`` :class:`FunctionDescriptor` instances."""
    path_lookup = _generate_long_paths(os_filter)

    # Change the treematch pattern into an fnmatch-able pattern to give back all functions from the sub-tree
    # (if there is a subtree).
    #
    # Examples:
    #     -f apps.webservers.iis -> apps.webservers.iis* (logs etc)
    #     -f apps.webservers.iis.logs -> apps.webservers.iis.logs* (only the logs, there is no subtree)
    # We do not include a dot because that does not work if the full path is given:
    #     -f apps.webservers.iis.logs != apps.webservers.iis.logs.* (does not work)
    search_pattern = pattern
    if not has_glob_magic(pattern):
        search_pattern += "*"

    result: list[FunctionDescriptor] = []
    for path in fnmatch.filter(path_lookup.keys(), search_pattern):
        for descriptor in path_lookup[path]:
            # Skip plugins that don't want to be found by wildcards
            if not descriptor.exported or (not show_hidden and not descriptor.findable):
                continue

            result.append(descriptor)

    yield from sorted(result, key=lambda desc: desc.module.count("."), reverse=True)


def _filter_compatible(
    descriptors: list[FunctionDescriptor], target: Target, ignore_load_errors: bool = False
) -> Iterator[FunctionDescriptor]:
    """Filter a list of function descriptors based on compatibility with a target."""
    compatible = set()
    incompatible = set()

    for descriptor in descriptors:
        if descriptor.qualname in compatible:
            yield descriptor
            continue

        if descriptor.qualname in incompatible:
            continue

        try:
            plugincls = load(descriptor)
        except Exception:
            if ignore_load_errors:
                incompatible.add(descriptor.qualname)
                continue
            raise

        try:
            if issubclass(plugincls, OSPlugin):
                if issubclass(target._os_plugin, plugincls):
                    compatible.add(descriptor.qualname)
                    yield descriptor
                else:
                    incompatible.add(descriptor.qualname)
                    continue
            elif plugincls(target).is_compatible():
                compatible.add(descriptor.qualname)
                yield descriptor
            else:
                incompatible.add(descriptor.qualname)
        except Exception:
            incompatible.add(descriptor.qualname)


def generate() -> dict[str, Any]:
    """Internal function to generate the list of available plugins.

    Walks the plugins directory and imports any ``.py`` files in there.
    Plugins will be automatically registered.

    Returns:
        The global ``PLUGINS`` dictionary.
    """
    plugins_dir = Path(__file__).parent / "plugins"
    for path in _find_py_files(plugins_dir):
        relative_path = path.relative_to(plugins_dir)
        module_tuple = (MODULE_PATH, *relative_path.parent.parts, relative_path.stem)
        load_module_from_name(".".join(module_tuple))

    return PLUGINS


def _find_py_files(path: Path) -> Iterator[Path]:
    """Walk all the files and directories in ``path`` and return all files ending in ``.py``.

    Do not walk or yield paths containing the following names:

    - __pycache__
    - __init__

    Furthermore, it logs an error if ``path`` does not exist.

    Args:
        path: The path to a directory or file to walk and filter.
    """
    if not path.exists():
        log.error("Path %s does not exist", path)
        return

    it = [path] if path.is_file() else path.glob("**/*.py")

    for entry in it:
        if not entry.is_file() or entry.name == "__init__.py":
            continue

        yield entry


def load_module_from_name(module_path: str) -> None:
    """Load a module from ``module_path``."""
    try:
        # This will trigger the __init__subclass__() of the Plugin subclasses in the module.
        importlib.import_module(module_path)
    except Exception as e:
        log.info("Unable to import %s", module_path)
        log.debug("Error while trying to import module %s", module_path, exc_info=e)
        _save_plugin_import_failure(module_path)


def load_module_from_file(path: Path, base_path: Path) -> None:
    """Loads a module from a file indicated by ``path`` relative to ``base_path``.

    The module is added to ``sys.modules`` so it can be found everywhere.

    Args:
        path: The file to load as module.
        base_path: The base directory of the module.
    """
    try:
        relative_path = path.relative_to(base_path)
        module_tuple = (*relative_path.parent.parts, relative_path.stem)
        spec = importlib.util.spec_from_file_location(".".join(module_tuple), path)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        sys.modules[module.__name__] = module
    except Exception as e:
        log.error("Unable to import %s", path)  # noqa: TRY400
        log.debug("Error while trying to import module %s", path, exc_info=e)
        _save_plugin_import_failure(str(path))


def load_modules_from_paths(paths: list[Path]) -> None:
    """Iterate over the ``paths`` and load all ``.py`` files."""
    for path in paths:
        for file in _find_py_files(path):
            base_path = path.parent if file == path else path
            load_module_from_file(file, base_path)


def get_external_module_paths(path_list: list[Path]) -> list[Path]:
    """Return a list of external plugin directories."""
    output_list = environment_variable_paths() + path_list

    return list(set(output_list))


def environment_variable_paths() -> list[Path]:
    """Return additional plugin directories specified by the ``DISSECT_PLUGINS`` environment variable."""
    env_var = os.environ.get("DISSECT_PLUGINS")

    plugin_dirs = env_var.split(":") if env_var else []

    return [Path(directory) for directory in plugin_dirs]


def _save_plugin_import_failure(module: str) -> None:
    """Store errors that occurred during plugin import."""
    stacktrace = traceback.format_exception(*sys.exc_info())
    PLUGINS.__failed__.append(FailureDescriptor(module, stacktrace))


def _get_nonprivate_attribute_names(cls: type[Plugin]) -> list[str]:
    """Retrieve all attributes that do not start with ``_``."""
    return [attr for attr in dir(cls) if not attr.startswith("_")]


def _get_nonprivate_attributes(cls: type[Plugin]) -> list[Any]:
    """Retrieve all public attributes of a :class:`Plugin`."""
    # Note: `dir()` might return attributes from parent class
    return [getattr(cls, attr) for attr in _get_nonprivate_attribute_names(cls)]


def _get_nonprivate_methods(cls: type[Plugin]) -> list[Callable]:
    """Retrieve all public methods of a :class:`Plugin`."""
    return [attr for attr in _get_nonprivate_attributes(cls) if not isinstance(attr, property) and callable(attr)]


def _get_descriptors_on_nonprivate_methods(cls: type[Plugin]) -> list[RecordDescriptor]:
    """Return record descriptors set on nonprivate methods in ``cls`` class."""
    descriptors = set()

    for m in _get_nonprivate_methods(cls):
        descriptors.update(_get_descriptors_on_method(m))

    return list(descriptors)


def _get_descriptors_on_method(method: Callable) -> list[RecordDescriptor]:
    """Return record descriptors set on a method."""
    if not (record := getattr(method, "__record__", None)):
        return []

    try:
        # check if __record__ value is iterable (for example, a list)
        return list(record)
    except TypeError:
        return [record]


# Classes for specific types of plugins
# These need to be at the bottom of the module because __init_subclass__ requires everything
# in the parent class Plugin to be defined and resolved.
class OSPlugin(Plugin):
    """Base class for OS plugins.

    This provides a base class for certain common functions of OS's, which each OS plugin has to implement separately.

    For example, it provides an interface for retrieving the hostname and users of a target.

    All derived classes MUST implement ALL the classmethods and exported
    methods with the same ``@classmethod`` or ``@export(...)`` annotation.
    """

    def __init_subclass__(cls, **kwargs):
        # Note that cls is the subclass
        super().__init_subclass__(**kwargs)

        for os_method in _get_nonprivate_attributes(OSPlugin):
            if isinstance(os_method, property):
                os_method = os_method.fget
            os_docstring = os_method.__doc__

            method = getattr(cls, os_method.__name__, None)
            if isinstance(method, property):
                method = method.fget
            # This works as None has a __doc__ property (which is None).
            docstring = method.__doc__

            if method and not docstring:
                if hasattr(method, "__func__"):
                    method = method.__func__
                method.__doc__ = os_docstring

    def check_compatible(self) -> bool:
        """OSPlugin's use a different compatibility check, override the one from the :class:`Plugin` class.

        Returns:
            This function always returns ``True``.
        """
        return True

    @classmethod
    def detect(cls, fs: Filesystem) -> Filesystem | None:
        """Provide detection of this OSPlugin on a given filesystem.

        Args:
            fs: :class:`~dissect.target.filesystem.Filesystem` to detect the OS on.

        Returns:
            The root filesystem / sysvol when found.
        """
        raise NotImplementedError

    @classmethod
    def create(cls, target: Target, sysvol: Filesystem) -> Self:
        """Initiate this OSPlugin with the given target and detected filesystem.

        Args:
            target: The :class:`~dissect.target.target.Target` object.
            sysvol: The filesystem that was detected in the ``detect()`` function.

        Returns:
            An instantiated version of the OSPlugin.
        """
        raise NotImplementedError

    @export(property=True)
    def hostname(self) -> str | None:
        """Return the target's hostname.

        Returns:
            The hostname as string.
        """
        raise NotImplementedError

    @export(property=True)
    def ips(self) -> list[str]:
        """Return the IP addresses configured in the target.

        Returns:
            The IPs as list.
        """
        raise NotImplementedError

    @export(property=True)
    def version(self) -> str | None:
        """Return the target's OS version.

        Returns:
            The OS version as string.
        """
        raise NotImplementedError

    @export(record=EmptyRecord)
    def users(self) -> list[Record]:
        """Return the users available in the target.

        Returns:
            A list of user records.
        """
        raise NotImplementedError

    @export(property=True)
    def os(self) -> str:
        """Return a slug of the target's OS name.

        Returns:
            A slug of the OS name, e.g. 'windows' or 'linux'.
        """
        raise NotImplementedError

    @internal
    def misc_user_paths(self) -> Iterator[tuple[str, tuple[str, str] | None]]:
        """Yields miscellaneous user paths and user keys.

        Example:
            ("c:/Windows/ServiceProfiles/LocalService", ("sid", "S-1-5-19"))
        """

        # Consider moving this to the UsersPlugin class, and create separate UsersPlugin subclasses for each OS
        # when the machinery concerning users and user_paths becomes more complex.
        raise NotImplementedError

    @internal
    def os_tree(self) -> list[str]:
        """Returns the :func:`os` value of this and all the OS plugin parents."""
        result: list[str] = []
        for klass in self.__class__.mro():
            if not issubclass(klass, OSPlugin):
                # In case the plugin extends multiple different classes.
                continue

            if klass is OSPlugin:
                break

            result.append(klass.os.__get__(self))
        # dicts maintain insertion order, we use it here to get rid of duplicates
        return list(dict.fromkeys(result))

    @export(property=True)
    def architecture(self) -> str | None:
        """Return a slug of the target's OS architecture.

        Returns:
            A slug of the OS architecture, e.g. 'x86_32-unix', 'MIPS-linux' or
            'AMD64-win32', or 'unknown' if the architecture is unknown.
        """
        raise NotImplementedError


class ChildTargetPlugin(Plugin):
    """A Child target is a special plugin that can list more Targets.

    For example, :class:`~dissect.target.plugins.child.esxi.ESXiChildTargetPlugin` can
    list all of the Virtual Machines on the host.
    """

    __type__ = None

    def list_children(self) -> Iterator[ChildTargetRecord]:
        """Yield :class:`~dissect.target.helpers.record.ChildTargetRecord` records of all
        possible child targets on this target.
        """
        raise NotImplementedError


class NamespacePlugin(Plugin):
    """A namespace plugin provides services to access functionality from a group of subplugins.

    Support is currently limited to shared exported functions with output type ``record`` and ``yield``.
    """

    def __init_subclass__(cls, **kwargs):
        # Upon subclassing, decide whether this is a direct subclass of NamespacePlugin
        # If this is not the case, autogenerate aggregate methods for methods record output.

        # Do not process the "base" subclasses defined in this file
        if cls.__module__ == NamespacePlugin.__module__:
            super().__init_subclass__(**kwargs)
            return

        if cls.__bases__[0] not in (NamespacePlugin, InternalNamespacePlugin):
            # This is a subclass of a subclass of (Internal)NamespacePlugin
            # We need to aggregate the methods of this subclass
            cls.__findable__ = True
            # Run the normal subclass initialization first so the plugin gets all its exports registered
            super().__init_subclass__(**kwargs)
            cls.__init_subclass_subplugin__(cls, **kwargs)
            cls.__nsplugin__.__init_subclass__(**kwargs)
        else:
            # This is a direct subclass of (Internal)NamespacePlugin
            # Not much to do here, just do some initialization and register the plugin
            cls.__findable__ = False
            super().__init_subclass__(**kwargs)

            cls.__nsplugin__ = cls
            if not hasattr(cls, "__subplugins__"):
                cls.__subplugins__ = set()

    def __init_subclass_subplugin__(cls, **kwargs) -> None:
        # Register the current plugin class as a subplugin with
        # the direct subclass of NamespacePlugin
        cls.__nsplugin__.__subplugins__.add(cls.__namespace__)

        # Generate a tuple of class names for which we do not want to add subplugin functions, which is the
        # NamespacePlugin and all of its superclasses (minus the base object)
        reserved_cls = tuple({klass.__name__ for klass in cls.__nsplugin__.mro() if klass is not object})

        # Collect the public exports of the subplugin
        for export_name in cls.__exports__:
            export_attr = getattr(cls, export_name)

            # The export need to be callable and may not be part of a parent class
            if not isinstance(export_attr, Callable) or export_attr.__qualname__.startswith(reserved_cls):
                continue

            # The export needs to output records or yield something else
            if (output_type := export_attr.__output__) not in ["record", "yield"]:
                continue

            if aggregator := getattr(cls.__nsplugin__, export_name, None):
                # If we already have an aggregate method, update some fields
                if not hasattr(aggregator, "__subplugins__"):
                    # This is not an aggregator, but a re-implementation of a subclass method by the subplugin
                    continue

                if aggregator.__output__ != output_type:
                    # The output type of the existing aggregator is different, so we can't merge them
                    raise PluginError(
                        "Cannot merge namespace methods with different output types"
                        f" ({aggregator.__output__} != {output_type} in {cls.__namespace__}.{export_name})"
                    )

                # Update record descriptors
                if (export_attr.__record__, aggregator.__record__) != (None, None):
                    aggregator.__record__ = list(
                        set(_get_descriptors_on_method(export_attr) + _get_descriptors_on_method(aggregator))
                    )
            else:
                # Generate a method for the parent namespace class
                aggregator = export(
                    output=output_type,
                    record=export_attr.__record__,
                    cache=False,
                    cls=cls.__nsplugin__,
                )(cls.__generate_aggregator(cls.__nsplugin__, export_name))

                # Add to the parent namespace class
                setattr(cls.__nsplugin__, export_name, aggregator)

            # Add subplugin to aggregator
            aggregator.__subplugins__.append(cls.__namespace__)
            cls.__update_aggregator_docs(aggregator)

    def check_compatible(self) -> None:
        at_least_one = False
        for entry in self.__subplugins__:
            try:
                self.target.get_function(entry)
                at_least_one = True
            except UnsupportedPluginError:  # noqa: PERF203
                self.target.log.warning("Subplugin %s is not compatible with target", entry)
            except Exception as e:
                self.target.log.error("Failed to load subplugin: %s", entry)  # noqa: TRY400
                self.target.log.debug("", exc_info=e)

        if not at_least_one:
            raise UnsupportedPluginError("No compatible subplugins found")

    @staticmethod
    def __generate_aggregator(nsklass: type[NamespacePlugin], method_name: str) -> Callable:
        def aggregator(self: NamespacePlugin) -> Iterator[Record]:
            for ns in aggregator.__subplugins__:
                try:
                    yield from self.target.get_function(f"{ns}.{method_name}")[1]()
                except UnsupportedPluginError:  # noqa: PERF203
                    continue
                except Exception as e:
                    self.target.log.error("Subplugin %s.%s raised an exception: %s", ns, method_name, e)  # noqa: TRY400
                    self.target.log.debug("", exc_info=e)

        # Holds the subplugins that share this method
        aggregator.__subplugins__ = []
        aggregator.__generated__ = True

        # Clone some names and attributes
        aggregator.__module__ = nsklass.__module__
        aggregator.__name__ = method_name
        aggregator.__qualname__ = f"{nsklass.__name__}.{method_name}"

        return aggregator

    @staticmethod
    def __update_aggregator_docs(aggregator: Callable) -> None:
        aggregator.__doc__ = f"Return {aggregator.__name__} for: {', '.join(aggregator.__subplugins__)}"


__INTERNAL_PLUGIN_METHOD_NAMES__ = {attr.__name__ for attr in _get_nonprivate_methods(Plugin)}


class InternalPlugin(Plugin):
    """Parent class for internal plugins.

    InternalPlugin marks all non-private methods internal by default
    (same as ``@internal`` decorator).
    """

    def __init_subclass__(cls, **kwargs):
        for method in _get_nonprivate_methods(cls):
            if callable(method) and method.__name__ not in __INTERNAL_PLUGIN_METHOD_NAMES__:
                method.__internal__ = True

        super().__init_subclass__(**kwargs)
        return cls


class InternalNamespacePlugin(NamespacePlugin, InternalPlugin):
    pass



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import ChildTargetRecord
from dissect.target.plugin import ChildTargetPlugin

if TYPE_CHECKING:
    from collections.abc import Iterator


class VirtuozzoChildTargetPlugin(ChildTargetPlugin):
    """Child target plugin that yields from Virtuozzo container's root.

    Virtuozzo containers are by default registered in the folder ``vz/root/$VEID``,
    where VEID will be substituted with the actual container UUID.

    .. code-block::

        /
        etc/
        var/
        vz/
            root/
                <container-uuid>/
                <container-uuid>/

    References:
        - https://docs.virtuozzo.com/virtuozzo_hybrid_server_7_command_line_reference/managing-system/configuration-files.html
        - https://wiki.openvz.org/Man/ctid.conf.5
        - https://docs.virtuozzo.com/pdf/virtuozzo_hybrid_server_7_command_line_reference.pdf
    """

    __type__ = "virtuozzo"

    PATH = "/vz/root"
    CONFIG_PATH = "/etc/vz/conf"

    def check_compatible(self) -> None:
        if not self.target.fs.path(self.PATH).exists():
            raise UnsupportedPluginError("No Virtuozzo path found")

    def list_children(self) -> Iterator[ChildTargetRecord]:
        for container in self.target.fs.path(self.PATH).iterdir():
            name = None
            try:
                if (vm_config := self.target.fs.path(self.CONFIG_PATH).joinpath(f"{container.name}.conf")).exists():
                    with vm_config.open("rt") as fh:
                        for line in fh:
                            if not (line := line.strip()):
                                continue

                            key, _, value = line.partition("=")
                            if key == "NAME":
                                name = value.strip('"')
                                break
            except Exception as e:
                self.target.log.exception("Failed to parse NAME from Virtuozzo config: %s", container)
                self.target.log.debug("", exc_info=e)

            yield ChildTargetRecord(
                type=self.__type__,
                name=name,
                path=container,
                _target=self.target,
            )



from __future__ import annotations

import io
import itertools
from itertools import product
from typing import TYPE_CHECKING, Final

from dissect.target.exceptions import ConfigurationParsingError, UnsupportedPluginError
from dissect.target.helpers.configutil import Default, ListUnwrapper, _update_dictionary
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import OperatingSystem, Plugin, arg, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.helpers import fsutil
    from dissect.target.target import Target

COMMON_ELEMENTS = [
    ("string", "name"),  # basename of .conf file
    ("string", "proto"),
    ("string", "dev"),
    ("string", "ca"),
    ("string", "cert"),
    ("string", "key"),
    ("boolean", "redacted_key"),
    ("string", "tls_auth"),
    ("string", "status"),
    ("string", "log"),
    ("string", "source"),
]

OpenVPNServer = TargetRecordDescriptor(
    "application/vpn/openvpn/server",
    [
        ("net.ipaddress", "local"),
        ("uint16", "port"),
        ("string", "dh"),
        ("string", "topology"),
        ("string", "server"),
        ("string", "ifconfig_pool_persist"),
        ("string[]", "pushed_options"),
        ("boolean", "client_to_client"),
        ("boolean", "duplicate_cn"),
        *COMMON_ELEMENTS,
    ],
)

OpenVPNClient = TargetRecordDescriptor(
    "application/vpn/openvpn/client",
    [
        ("string[]", "remote"),
        *COMMON_ELEMENTS,
    ],
)


class OpenVPNParser(Default):
    def __init__(self, *args, **kwargs):
        boolean_fields = OpenVPNServer.getfields("boolean") + OpenVPNClient.getfields("boolean")
        self.boolean_field_names = {field.name.replace("_", "-") for field in boolean_fields}

        super().__init__(*args, separator=(r"\s",), collapse=["key", "ca", "cert"], **kwargs)

    def parse_file(self, fh: io.TextIOBase) -> None:
        root = {}
        iterator = self.line_reader(fh)
        for line in iterator:
            if line.startswith("<"):
                key = line.strip().strip("<>")
                value = self._read_blob(iterator)
                _update_dictionary(root, key, value)
                continue

            self._parse_line(root, line)

        self.parsed_data = ListUnwrapper.unwrap(root)

    def _read_blob(self, lines: Iterator[str]) -> str | list[dict]:
        """Read the whole section between ``<data></data>`` sections."""
        output = ""
        with io.StringIO() as buffer:
            for line in lines:
                if "</" in line:
                    break

                buffer.write(line)
            output = buffer.getvalue()

        # Check for connection profile blocks
        if not output.startswith("-----"):
            profile_dict = {}
            for line in output.splitlines():
                self._parse_line(profile_dict, line)

            # We put it as a list as _update_dictionary appends data in a list.
            output = [profile_dict]

        return output

    def _parse_line(self, root: dict, line: str) -> None:
        key, *value = self.SEPARATOR.split(line, 1)
        # Unquote data
        value = value[0].strip() if value else ""

        value = value.strip("'\"")

        if key in self.boolean_field_names:
            value = True

        _update_dictionary(root, key, value)


class OpenVPNPlugin(Plugin):
    """OpenVPN configuration parser.

    References:
        - man (8) openvpn
    """

    __namespace__ = "openvpn"

    config_globs = (
        # This catches openvpn@, openvpn-client@, and openvpn-server@ systemd configurations
        # Linux
        "/etc/openvpn/",
        # Windows
        "sysvol/Program Files/OpenVPN/config/",
    )

    user_config_paths: Final[dict[str, list[str]]] = {
        OperatingSystem.WINDOWS.value: ["OpenVPN/config/"],
        OperatingSystem.OSX.value: ["Library/Application Support/OpenVPN Connect/profiles/"],
    }

    def __init__(self, target: Target):
        super().__init__(target)
        self.configs: list[fsutil.TargetPath] = []
        for base, glob in product(self.config_globs, ["*.conf", "*.ovpn"]):
            self.configs.extend(self.target.fs.path(base).rglob(glob))

        user_paths = self.user_config_paths.get(target.os, [])
        for path, glob, user_details in itertools.product(
            user_paths, ["*.conf", "*.ovpn"], self.target.user_details.all_with_home()
        ):
            self.configs.extend(user_details.home_path.joinpath(path).rglob(glob))

    def check_compatible(self) -> None:
        if not self.configs:
            raise UnsupportedPluginError("No OpenVPN configuration files found")

    def _load_config(self, parser: OpenVPNParser, config_path: fsutil.TargetPath) -> dict | None:
        with config_path.open("rt") as file:
            try:
                parser.parse_file(file)
            except ConfigurationParsingError as e:
                # Couldn't parse file, continue
                self.target.log.info("An issue occurred during parsing of %s, continuing", config_path)
                self.target.log.debug("", exc_info=e)
                return None

        return parser.parsed_data

    @export(record=[OpenVPNServer, OpenVPNClient])
    @arg("--export-key", action="store_true", help="export private keys to records")
    def config(self, export_key: bool = False) -> Iterator[OpenVPNServer | OpenVPNClient]:
        """Parses config files from openvpn interfaces."""
        # We define the parser here so we can reuse it
        parser = OpenVPNParser()

        for config_path in self.configs:
            config = self._load_config(parser, config_path)

            common_elements = {
                "name": config_path.stem,
                "proto": config.get("proto", "udp"),  # Default is UDP
                "dev": config.get("dev"),
                "ca": config.get("ca"),
                "cert": config.get("cert"),
                "key": config.get("key"),
                "status": config.get("status"),
                "log": config.get("log"),
                "source": config_path,
                "_target": self.target,
            }

            if not export_key and "PRIVATE KEY" in common_elements.get("key"):
                common_elements.update({"key": None})
                common_elements.update({"redacted_key": True})

            tls_auth = config.get("tls-auth", "")
            # The format of tls-auth is 'tls-auth ta.key <NUM>'.
            # NUM is either 0 or 1 depending on whether the configuration
            # is for the client or server, and that does not interest us
            # This gets rid of the number at the end, while still supporting spaces
            tls_auth = " ".join(tls_auth.split(" ")[:-1]).strip("'\"")

            common_elements.update({"tls_auth": tls_auth})

            if "client" in config:
                remote = config.get("remote", [])

                yield OpenVPNClient(
                    **common_elements,
                    remote=remote,
                )
            else:
                # Defaults here are taken from `man (8) openvpn`
                yield OpenVPNServer(
                    **common_elements,
                    local=config.get("local", "0.0.0.0"),
                    port=int(config.get("port", "1194")),
                    dh=config.get("dh"),
                    topology=config.get("topology"),
                    server=config.get("server"),
                    ifconfig_pool_persist=config.get("ifconfig-pool-persist"),
                    pushed_options=config.get("push", []),
                    client_to_client=config.get("client-to-client", False),
                    duplicate_cn=config.get("duplicate-cn", False),
                )



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import ChildTargetRecord
from dissect.target.plugin import ChildTargetPlugin

if TYPE_CHECKING:
    from collections.abc import Iterator


class ProxmoxChildTargetPlugin(ChildTargetPlugin):
    """Child target plugin that yields from the VM listing."""

    __type__ = "proxmox"

    def check_compatible(self) -> None:
        if self.target.os != "proxmox":
            raise UnsupportedPluginError("Not a Proxmox operating system")

    def list_children(self) -> Iterator[ChildTargetRecord]:
        for vm in self.target.vmlist():
            vm_path = self.target.fs.path(vm.path)

            name = None
            try:
                with vm_path.open("rt") as fh:
                    for line in fh:
                        if not (line := line.strip()):
                            continue

                        if (key_value := line.split(":", 1)) and key_value[0] == "name":
                            name = key_value[1].strip()
                            break
            except Exception as e:
                self.target.log.error("Failed parsing name from VM config: %s", vm_path)  # noqa: TRY400
                self.target.log.debug("", exc_info=e)

            yield ChildTargetRecord(
                type=self.__type__,
                name=name,
                path=vm_path,
                _target=self.target,
            )



from __future__ import annotations

import re
from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.utils import year_rollover_helper
from dissect.target.plugin import export
from dissect.target.plugins.os.unix.packagemanager import (
    OperationTypes,
    PackageManagerLogRecord,
    PackageManagerPlugin,
)

if TYPE_CHECKING:
    from collections.abc import Iterator

YUM_LOG_KEYWORDS = ["Installed", "Updated", "Erased", "Obsoleted"]
RE_TS = re.compile(r"(\w+\s{1,2}\d+\s\d{2}:\d{2}:\d{2})")


class YumPlugin(PackageManagerPlugin):
    """Yum package manager plugin."""

    __namespace__ = "yum"

    LOG_DIR_PATH = "/var/log"
    LOG_FILES_GLOB = "yum.*"

    def check_compatible(self) -> None:
        if not next(self.target.fs.path(self.LOG_DIR_PATH).glob(self.LOG_FILES_GLOB), None):
            raise UnsupportedPluginError("No Yum files found")

    @export(record=PackageManagerLogRecord)
    def logs(self) -> Iterator[PackageManagerLogRecord]:
        """Package manager log parser for CentOS' Yellowdog Updater (Yum).

        Example log format::

            Dec 16 04:41:22 Installed: unzip-6.0-24.el7_9.x86_64
            Dec 16 04:41:25 Installed: unzip-6.0-22.el7_9.x86_64
            Dec 16 04:41:28 Updated: unzip-6.0-24.el7_9.x86_64
            Dec 16 04:41:30 Erased: unzip-6.0-24.el7_9.x86_64
            Dec 16 04:41:34 Installed: unzip-6.0-24.el7_9.x86_64
        """
        target_tz = self.target.datetime.tzinfo

        for path in self.target.fs.path(self.LOG_DIR_PATH).glob(self.LOG_FILES_GLOB):
            for ts, line in year_rollover_helper(path, RE_TS, "%b %d %H:%M:%S", target_tz):
                # Only parse lines that are about installation/erasions/updates, not empty lines or debug statements.
                if not any(keyword in line for keyword in YUM_LOG_KEYWORDS):
                    return

                operation, package_name = line[16:].split(": ")
                yield PackageManagerLogRecord(
                    ts=ts,
                    package_manager="yum",
                    operation=OperationTypes.infer(operation.strip()).value,
                    package_name=package_name,
                    _target=self.target,
                )



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from dissect.target.helpers.regutil import VirtualHive, VirtualKey, VirtualValue
from dissect.target.plugins.os.windows.ad.ntds import DEFAULT_NT_HASH
from tests._utils import absolute_path
from tests.plugins.os.windows.credential.test_credhist import md4
from tests.plugins.os.windows.test_lsa import map_lsa_system_keys

if TYPE_CHECKING:
    from dissect.target.target import Target


@pytest.fixture
def target_win_ntds(target_win: Target, hive_hklm: VirtualHive) -> Target:
    registry_path = "SYSTEM\\ControlSet001\\Services\\NTDS\\Parameters"
    hive_hklm.map_key(registry_path, VirtualKey(hive_hklm, registry_path))
    hive_hklm.map_value(
        registry_path,
        "DSA Database file",
        VirtualValue(hive_hklm, "DSA Database file", "c:/windows/ntds/ntds.dit"),
    )

    map_lsa_system_keys(
        hive_hklm,
        {
            "JD": "ebaa656d",
            "Skew1": "959f28b0",
            "GBG": "0766a85b",
            "Data": "1af1b31e",
        },
    )

    target_win.fs.map_file(
        "c:/windows/ntds/ntds.dit",
        absolute_path("_data/plugins/os/windows/ad/ntds/goad/ntds.dit.gz"),
        compression="gzip",
    )

    return target_win


def test_users(target_win_ntds: Target) -> None:
    """Tests if ``ad.users`` outputs the correct amount of records and their content"""
    cn_to_ntlm_hash_mapping = {
        "krbtgt": "988160b622eb37838dbff2523015e44c",  # Unknown Password
        "NORTH$": "8048b2621bb71945d6ca6e9a14084af1",  # Unknown Password
        "ESSOS$": "f1580437d0120689ad3909b9fe9b74fe",  # Unknown Password
        "Administrator": "c66d72021a2d4744409969a581a1705e",  # Unknown Password
        "renly.baratheon": "f667bd83b30c87801cef53856618d534",  # Unknown Password
        "vagrant": md4("vagrant").hex(),
        "lord.varys": md4("_W1sper_$").hex(),
        "jaime.lannister": md4("cersei").hex(),
        "tyron.lannister": md4("Alc00L&S3x").hex(),
        "cersei.lannister": md4("il0vejaime").hex(),
        "joffrey.baratheon": md4("1killerlion").hex(),
        "stannis.baratheon": md4("Drag0nst0ne").hex(),
        "petyer.baelish": md4("@littlefinger@").hex(),
        "tywin.lannister": md4("powerkingftw135").hex(),
        "maester.pycelle": md4("MaesterOfMaesters").hex(),
    }

    results = list(target_win_ntds.ad.users())

    assert len(results) == 33

    for result in results:
        if result.cn not in cn_to_ntlm_hash_mapping or result.nt == DEFAULT_NT_HASH:
            continue

        assert cn_to_ntlm_hash_mapping[result.cn] == result.nt


def test_computers(target_win_ntds: Target) -> None:
    """Tests if ``ad.computers`` outputs the correct amount of records and their content"""
    cn_to_ntlm_hash_mapping = {
        "KINGSLANDING": "00e3201a59af7ecc72e939a8c9794c64",  # Unknown Password
    }

    results = list(target_win_ntds.ad.computers())

    assert len(results) == 3

    for result in results:
        if result.cn not in cn_to_ntlm_hash_mapping or result.nt == DEFAULT_NT_HASH:
            continue

        assert cn_to_ntlm_hash_mapping[result.cn] == result.nt


def test_group_policies(target_win_ntds: Target) -> None:
    results = list(target_win_ntds.ad.group_policies())

    assert len(results) == 5



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import TYPE_CHECKING
from unittest.mock import MagicMock, Mock, patch

import pytest

from dissect.target.helpers.regutil import VirtualHive, VirtualKey
from dissect.target.plugins.os.windows._os import WindowsPlugin
from dissect.target.plugins.os.windows.network import WindowsNetworkPlugin
from tests.conftest import change_controlset

if TYPE_CHECKING:
    from dissect.target.target import Target


@dataclass
class MockRegVal:
    name: str
    value: str | int


REGISTRY_KEY_INTERFACE = "HKLM\\SYSTEM\\CurrentControlSet\\Services\\Tcpip\\Parameters\\Interfaces\\"
REGISTRY_KEY_CONTROLSET = "HKLM\\SYSTEM\\CurrentControlSet\\Control\\Class\\{4d36e972-e325-11ce-bfc1-08002be10318}"
REGISTRY_KEY_CONNECTION = "HKLM\\SYSTEM\\CurrentControlSet\\Control\\Network\\{4D36E972-E325-11CE-BFC1-08002BE10318}\\"


@pytest.mark.parametrize(
    ("mock_values", "expected_values"),
    [
        pytest.param(
            {
                "VlanID": 12,
                "NetCfgInstanceId": "TESTINGINSTANCEID",
                "*IfType": 1,
                "NetworkAddress": "DE:AD:BE:EF:DE:AD",
                "NetworkInterfaceInstallTimestamp": 130005216000000000,
                "DhcpDefaultGateway": ["10.10.10.1"],
                "DhcpIPAddress": "10.10.10.10",
                "DhcpNameServer": "10.10.10.2",
                "DhcpSubnetMask": "255.255.255.0",
                "InterfaceMetric": 15,
                "DhcpDomain": "work",
                "Name": "ETHERNET 0",
                "EnableDHCP": 1,
            },
            {
                "name": "ETHERNET 0",
                "type": "OTHER",
                "enabled": True,
                "cidr": ["10.10.10.10/24"],
                "ip": ["10.10.10.10"],
                "gateway": ["10.10.10.1"],
                "dns": ["10.10.10.2"],
                "mac": ["DE:AD:BE:EF:DE:AD"],
                "metric": 15,
                "search_domain": ["work"],
                "first_connected": datetime.fromisoformat("2012-12-21 00:00:00+00:00"),
                "dhcp": True,
                "vlan": 12,
            },
            id="DHCP enabled",
        ),
        pytest.param(
            {
                "VlanID": 11,
                "NetCfgInstanceId": "TESTINGINSTANCEID",
                "*IfType": 6,
                "NetworkAddress": "DE:AD:BE:EF:DE:AD",
                "NetworkInterfaceInstallTimestamp": 130005216000000000,
                "DefaultGateway": ["10.10.10.1"],
                "IPAddress": "10.10.10.10",
                "NameServer": "10.10.10.2",
                "SubnetMask": "255.255.255.0",
                "InterfaceMetric": 5,
                "Domain": "local",
                "EnableDHCP": 0,
            },
            {
                "type": "ETHERNET_CSMACD",
                "cidr": ["10.10.10.10/24"],
                "ip": ["10.10.10.10"],
                "gateway": ["10.10.10.1"],
                "dns": ["10.10.10.2"],
                "mac": ["DE:AD:BE:EF:DE:AD"],
                "metric": 5,
                "search_domain": ["local"],
                "first_connected": datetime.fromisoformat("2012-12-21 00:00:00+00:00"),
                "dhcp": False,
                "vlan": 11,
            },
            id="DHCP disabled",
        ),
        pytest.param(
            {
                "NetCfgInstanceId": "TESTINGINSTANCEID",
                "*IfType": 1,
                "DefaultGateway": ["10.10.10.1"],
            },
            {
                "type": "OTHER",
                "gateway": ["10.10.10.1"],
                "dhcp": False,
            },
            id="OTHER",
        ),
        pytest.param(
            {
                "NetCfgInstanceId": "TESTINGINSTANCEID",
                "*IfType": 1,
                "DefaultGateway": ["10.10.10.2"],
                "NetworkAddress": None,
                "NetworkInterfaceInstallTimestamp": None,
                "DhcpIPAddress": None,
                "DhcpNameServer": None,
                "DhcpSubnetMask": None,
                "Domain": None,
                "VlanID": None,
                "InterfaceMetric": None,
                "Name": None,
                "EnableDHCP": None,
            },
            {
                "type": "OTHER",
                "gateway": ["10.10.10.2"],
                "search_domain": [],
                "dhcp": False,
            },
            id="OTHER empty",
        ),
        pytest.param(
            {
                "NetCfgInstanceId": "TESTINGINSTANCEID",
                "EnableDHCP": 1,
                "DhcpIPAddress": "0.0.0.0",
                "DhcpSubnetMask": "255.255.255.0",
            },
            {
                "_NO_INTERFACES": True,
                "enabled": True,
                "cidr": {},  # and not {'/255.255.255.0'}
            },
            id="invalid cidr",
        ),
    ],
)
def test_windows_network(
    mock_values: dict[str, MockRegVal],
    expected_values: dict[str, str | int],
    target_win: Target,
) -> None:
    mock_value_dict = {key: MockRegVal(name=key, value=value) for key, value in mock_values.items()}
    mock_registry = Mock()
    mock_registry.keys.return_value = [mock_registry]
    mock_registry.key.side_effect = lambda name: (
        mock_registry
        if name
        in [
            f"{REGISTRY_KEY_CONNECTION}TESTINGINSTANCEID\\Connection",
            f"{REGISTRY_KEY_INTERFACE}TESTINGINSTANCEID",
        ]
        else None
    )
    mock_registry.__len__ = lambda self: len(mock_value_dict)
    mock_registry.subkeys.return_value = [mock_registry]
    mock_registry.value.side_effect = lambda field: mock_value_dict.get(field, MockRegVal(name=field, value=None))
    mock_registry.values.return_value = list(mock_value_dict.values())

    with (
        patch("dissect.target.plugins.os.windows.generic.GenericPlugin", return_value=""),
        patch(
            "dissect.target.plugins.os.windows._os.WindowsPlugin.hostname", property(MagicMock(return_value="hostname"))
        ),
        patch.object(target_win, "registry", mock_registry),
    ):
        network = target_win.network
        assert network.ips() == expected_values.get("ip", [])
        assert network.dns() == expected_values.get("dns", [])
        assert network.gateways() == expected_values.get("gateway", [])
        assert network.macs() == expected_values.get("mac", [])

        if expected_values.get("_NO_INTERFACES"):
            with pytest.raises(StopIteration):
                next(iter(network.interfaces()))
            return

        network_interface = next(iter(network.interfaces()))
        assert network_interface.name == expected_values.get("name")
        assert network_interface.type == expected_values.get("type")
        assert network_interface.enabled == expected_values.get("enabled")
        assert network_interface.cidr == expected_values.get("cidr", [])
        assert network_interface.metric == expected_values.get("metric")
        assert network_interface.search_domain == expected_values.get("search_domain", [])
        assert network_interface.first_connected == expected_values.get("first_connected")
        assert network_interface.dhcp == expected_values.get("dhcp")
        assert network_interface.vlan == expected_values.get("vlan")


@pytest.mark.parametrize(
    "mock_values",
    [
        {
            "NetCfgInstanceId": "TESTINGINSTANCEID",
            "*IfType": None,
        },
    ],
)
def test_windows_network_none(
    mock_values: dict[str, MockRegVal],
    target_win: Target,
) -> None:
    mock_value_dict = {key: MockRegVal(name=key, value=value) for key, value in mock_values.items()}
    mock_registry = Mock()
    mock_registry.keys.return_value = [mock_registry]
    mock_registry.key.side_effect = lambda name: (
        mock_registry
        if name
        in [
            f"{REGISTRY_KEY_CONNECTION}TESTINGINSTANCEID\\Connection",
            f"{REGISTRY_KEY_INTERFACE}TESTINGINSTANCEID",
        ]
        else None
    )
    mock_registry.__len__ = lambda self: len(mock_value_dict)
    mock_registry.subkeys.return_value = [mock_registry]
    mock_registry.value.side_effect = lambda field: mock_value_dict.get(field, MockRegVal(name=field, value=None))
    mock_registry.values.return_value = list(mock_value_dict.values())

    with (
        patch(
            "dissect.target.plugins.os.windows._os.WindowsPlugin.hostname", property(MagicMock(return_value="hostname"))
        ),
        patch.object(target_win, "registry", mock_registry),
    ):
        network = target_win.network
        network_interface = list(network.interfaces())
        assert network.ips() == []
        assert network.dns() == []
        assert network.gateways() == []
        assert network.macs() == []
        assert network_interface == []


@pytest.mark.parametrize(
    ("mock_values", "expected_values"),
    [
        (
            {
                "NetCfgInstanceId": "TESTINGINSTANCEID",
                "*IfType": 24,
                "NetworkAddress": "FE:EE:EE:EE:EE:ED",
                "NetworkInterfaceInstallTimestamp": 130005216000000000,
                "DhcpDefaultGateway": ["192.168.0.1"],
                "DefaultGateway": ["10.0.0.1"],
                "DhcpIPAddress": "192.168.0.10",
                "IPAddress": "10.0.0.10",
                "DhcpNameServer": "192.168.0.2",
                "NameServer": "10.0.0.2 10.0.0.3",
                "SubnetMask": "255.255.255.0",
                "DhcpSubnetMask": "255.255.255.0",
                "VlanID": 10,
                "InterfaceMetric": 20,
                "Domain": "corp",
                "DhcpDomain": "corp",
                "EnableDHCP": 1,
            },
            [
                {
                    "name": None,
                    "type": "SOFTWARE_LOOPBACK",
                    "enabled": True,
                    "cidr": ["192.168.0.10/24"],
                    "ip": ["192.168.0.10"],
                    "gateway": ["192.168.0.1"],
                    "dns": ["192.168.0.2"],
                    "mac": ["FE:EE:EE:EE:EE:ED"],
                    "metric": 20,
                    "search_domain": ["corp"],
                    "first_connected": datetime.fromisoformat("2012-12-21 00:00:00+00:00"),
                    "dhcp": True,
                    "vlan": 10,
                },
                {
                    "name": None,
                    "type": "SOFTWARE_LOOPBACK",
                    "enabled": None,
                    "cidr": ["10.0.0.10/24"],
                    "ip": ["10.0.0.10"],
                    "gateway": ["10.0.0.1"],
                    "dns": ["10.0.0.2", "10.0.0.3"],
                    "mac": ["FE:EE:EE:EE:EE:ED"],
                    "metric": 20,
                    "search_domain": ["corp"],
                    "first_connected": datetime.fromisoformat("2012-12-21 00:00:00+00:00"),
                    "dhcp": False,
                    "vlan": 10,
                },
            ],
        ),
    ],
)
def test_network_dhcp_and_static(
    mock_values: dict[str, str | int | list[str] | None],
    expected_values: list[dict[str, str | int | list[str] | None]],
    target_win: Target,
) -> None:
    mock_value_dict = {key: MockRegVal(name=key, value=value) for key, value in mock_values.items()}
    mock_registry = Mock()
    mock_registry.keys.return_value = [mock_registry]
    mock_registry.key.side_effect = lambda name: (
        mock_registry
        if name
        in [
            f"{REGISTRY_KEY_CONNECTION}TESTINGINSTANCEID\\Connection",
            f"{REGISTRY_KEY_INTERFACE}TESTINGINSTANCEID",
        ]
        else None
    )
    mock_registry.__len__ = lambda self: len(mock_value_dict)
    mock_registry.subkeys.return_value = [mock_registry]
    mock_registry.value.side_effect = lambda field: mock_value_dict.get(field, MockRegVal(name=field, value=None))
    mock_registry.values.return_value = list(mock_value_dict.values())

    with (
        patch("dissect.target.plugins.os.windows.generic.GenericPlugin", return_value=""),
        patch(
            "dissect.target.plugins.os.windows._os.WindowsPlugin.hostname", property(MagicMock(return_value="hostname"))
        ),
        patch.object(target_win, "registry", mock_registry),
    ):
        network = target_win.network
        interfaces = list(network.interfaces())

        ips = set()
        dns = set()
        gateways = set()
        macs = set()

        for interface, expected in zip(interfaces, expected_values, strict=False):
            ips.update({iface.ip for iface in interface.cidr})
            dns.update(interface.dns)
            gateways.update(interface.gateway)
            macs.update(interface.mac)

            assert sorted(str(x.ip) for x in interface.cidr) == expected["ip"]
            assert sorted(map(str, interface.dns)) == expected["dns"]
            assert interface.name == expected["name"]
            assert interface.type == expected["type"]
            assert interface.enabled == expected["enabled"]
            assert interface.cidr == expected["cidr"]
            assert interface.gateway == expected["gateway"]
            assert interface.mac == expected["mac"]
            assert interface.first_connected == expected["first_connected"]
            assert interface.metric == expected["metric"]
            assert interface.search_domain == expected["search_domain"]
            assert interface.dhcp == expected["dhcp"]
            assert interface.vlan == expected["vlan"]

        assert network.ips() == list(ips)
        assert network.dns() == list(dns)
        assert network.gateways() == list(gateways)
        assert network.macs() == list(macs)


@patch(
    "dissect.target.plugins.os.windows.registry.RegistryPlugin.controlsets",
    property(MagicMock(return_value=["ControlSet001", "ControlSet002", "ControlSet003"])),
)
def test_regression_duplicate_ips(target_win: Target, hive_hklm: VirtualHive) -> None:
    """Regression test for https://github.com/fox-it/dissect.target/issues/877."""

    change_controlset(hive_hklm, 3)

    # register the interfaces
    kvs = [
        (
            "SYSTEM\\ControlSet001\\Control\\Class\\{4d36e972-e325-11ce-bfc1-08002be10318}\\0001",
            "{some-net-cfg-instance-uuid}",
        ),
        (
            "SYSTEM\\ControlSet002\\Control\\Class\\{4d36e972-e325-11ce-bfc1-08002be10318}\\0002",
            "{some-net-cfg-instance-uuid}",
        ),
        (
            "SYSTEM\\ControlSet003\\Control\\Class\\{4d36e972-e325-11ce-bfc1-08002be10318}\\0003",
            "{some-net-cfg-instance-uuid}",
        ),
    ]
    for name, value in kvs:
        key = VirtualKey(hive_hklm, name)
        key.add_value("NetCfgInstanceId", value)
        hive_hklm.map_key(name, key)

    # register interface dhcp ip addresses for three different control sets
    kvs = [
        ("SYSTEM\\ControlSet001\\Services\\Tcpip\\Parameters\\Interfaces\\{some-net-cfg-instance-uuid}", "1.2.3.4"),
        ("SYSTEM\\ControlSet002\\Services\\Tcpip\\Parameters\\Interfaces\\{some-net-cfg-instance-uuid}", "1.2.3.4"),
        ("SYSTEM\\ControlSet003\\Services\\Tcpip\\Parameters\\Interfaces\\{some-net-cfg-instance-uuid}", "5.6.7.8"),
    ]
    for name, value in kvs:
        key = VirtualKey(hive_hklm, name)
        key.add_value("DhcpIPAddress", value)
        hive_hklm.map_key(name, key)

    target_win.add_plugin(WindowsPlugin)
    target_win.add_plugin(WindowsNetworkPlugin)

    assert isinstance(target_win.ips, list)
    assert all(isinstance(ip, str) for ip in target_win.ips)
    assert len(target_win.ips) == 2
    assert sorted(target_win.ips) == ["1.2.3.4", "5.6.7.8"]



from __future__ import annotations

import datetime
import struct
from binascii import crc32
from io import BytesIO
from typing import TYPE_CHECKING, Union, get_args

from dissect.cstruct import cstruct
from dissect.util.compression import lznt1
from dissect.util.ts import wintimestamp

from dissect.target.exceptions import RegistryValueNotFoundError, UnsupportedPluginError
from dissect.target.helpers.descriptor_extensions import UserRecordDescriptorExtension
from dissect.target.helpers.record import (
    TargetRecordDescriptor,
    create_extended_descriptor,
)
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from flow.record import Record

    from dissect.target.target import Target

# References:
# - generaltel.dll
# - win32k.sys (Windows 7)
# - win32kbase.sys (Windows 10)
cit_def = """
typedef QWORD FILETIME;

flag TELEMETRY_ANSWERS {
    Unknown0        = 0x4,
    Unknown1        = 0x8,
    POWERBROADCAST  = 0x10000,
    DEVICECHANGE    = 0x20000,
    IME_CONTROL     = 0x40000,
    WINHELP         = 0x80000,
};

typedef struct _CIT_HEADER {
    WORD    MajorVersion;
    WORD    MinorVersion;
    DWORD   Size;                   /* Size of the entire buffer */
    FILETIME    CurrentTimeLocal;   /* Maybe the time when the saved CIT was last updated? */
    DWORD   Crc32;                  /* Crc32 of the entire buffer, skipping this field */
    DWORD   EntrySize;
    DWORD   EntryCount;
    DWORD   EntryDataOffset;
    DWORD   SystemDataSize;
    DWORD   SystemDataOffset;
    DWORD   BaseUseDataSize;
    DWORD   BaseUseDataOffset;
    FILETIME    StartTimeLocal;     /* Presumably when the aggregation started */
    FILETIME    PeriodStartLocal;   /* Presumably the starting point of the aggregation period */
    DWORD   AggregationPeriodInS;   /* Presumably the duration over which this data was gathered
                                     * Always 604800 (7 days) */
    DWORD   BitPeriodInS;           /* Presumably the amount of seconds a single bit represents
                                     * Always 3600 (1 hour) */
    DWORD   SingleBitmapSize;       /* This appears to be the sizes of the Stats buffers, always 21 */
    DWORD   _Unk0;                  /* Always 0x00000100? */
    DWORD   HeaderSize;
    DWORD   _Unk1;                  /* Always 0x00000000? */
} CIT_HEADER;

typedef struct _CIT_PERSISTED {
    DWORD   BitmapsOffset;          /* Array of Offset and Size (DWORD, DWORD) */
    DWORD   BitmapsSize;
    DWORD   SpanStatsOffset;        /* Array of Count and Duration (DWORD, DWORD) */
    DWORD   SpanStatsSize;
    DWORD   StatsOffset;            /* Array of WORD */
    DWORD   StatsSize;
} CIT_PERSISTED;

typedef struct _CIT_ENTRY {
    DWORD   ProgramDataOffset;      /* Offset to CIT_PROGRAM_DATA */
    DWORD   UseDataOffset;          /* Offset to CIT_PERSISTED */
    DWORD   ProgramDataSize;
    DWORD   UseDataSize;
} CIT_ENTRY;

typedef struct _CIT_PROGRAM_DATA {
    DWORD   FilePathOffset;         /* Offset to UTF-16-LE file path string */
    DWORD   FilePathSize;           /* strlen of string */
    DWORD   CommandLineOffset;      /* Offset to UTF-16-LE command line string */
    DWORD   CommandLineSize;        /* strlen of string */
    DWORD   PeTimeDateStamp;        /* aka Extra1 */
    DWORD   PeCheckSum;             /* aka Extra2 */
    DWORD   Extra3;                 /* aka Extra3, some flag from PROCESSINFO struct */
} CIT_PROGRAM_DATA;

typedef struct _CIT_BITMAP_ITEM {
    DWORD   Offset;
    DWORD   Size;
} CIT_BITMAP_ITEM;

typedef struct _CIT_SPAN_STAT_ITEM {
    DWORD   Count;
    DWORD   Duration;
} CIT_SPAN_STAT_ITEM;

typedef struct _CIT_SYSTEM_DATA_SPAN_STATS {
    CIT_SPAN_STAT_ITEM  ContextFlushes0;
    CIT_SPAN_STAT_ITEM  Foreground0;
    CIT_SPAN_STAT_ITEM  Foreground1;
    CIT_SPAN_STAT_ITEM  DisplayPower0;
    CIT_SPAN_STAT_ITEM  DisplayRequestChange;
    CIT_SPAN_STAT_ITEM  DisplayPower1;
    CIT_SPAN_STAT_ITEM  DisplayPower2;
    CIT_SPAN_STAT_ITEM  DisplayPower3;
    CIT_SPAN_STAT_ITEM  ContextFlushes1;
    CIT_SPAN_STAT_ITEM  Foreground2;
    CIT_SPAN_STAT_ITEM  ContextFlushes2;
} CIT_SYSTEM_DATA_SPAN_STATS;

typedef struct _CIT_USE_DATA_SPAN_STATS {
    CIT_SPAN_STAT_ITEM  ProcessCreation0;
    CIT_SPAN_STAT_ITEM  Foreground0;
    CIT_SPAN_STAT_ITEM  Foreground1;
    CIT_SPAN_STAT_ITEM  Foreground2;
    CIT_SPAN_STAT_ITEM  ProcessSuspended;
    CIT_SPAN_STAT_ITEM  ProcessCreation1;
} CIT_USE_DATA_SPAN_STATS;

typedef struct _CIT_SYSTEM_DATA_STATS {
    WORD    Unknown_BootIdRelated0;
    WORD    Unknown_BootIdRelated1;
    WORD    Unknown_BootIdRelated2;
    WORD    Unknown_BootIdRelated3;
    WORD    Unknown_BootIdRelated4;
    WORD    SessionConnects;
    WORD    ProcessForegroundChanges;
    WORD    ContextFlushes;
    WORD    MissingProgData;
    WORD    DesktopSwitches;
    WORD    WinlogonMessage;
    WORD    WinlogonLockHotkey;
    WORD    WinlogonLock;
    WORD    SessionDisconnects;
} CIT_SYSTEM_DATA_STATS;

typedef struct _CIT_USE_DATA_STATS {
    WORD    Crashes;
    WORD    ThreadGhostingChanges;
    WORD    Input;
    WORD    InputKeyboard;
    WORD    Unknown;
    WORD    InputTouch;
    WORD    InputHid;
    WORD    InputMouse;
    WORD    MouseLeftButton;
    WORD    MouseRightButton;
    WORD    MouseMiddleButton;
    WORD    MouseWheel;
} CIT_USE_DATA_STATS;

// PUU
typedef struct _CIT_POST_UPDATE_USE_INFO {
    DWORD   UpdateKey;
    WORD    UpdateCount;
    WORD    CrashCount;
    WORD    SessionCount;
    WORD    LogCount;
    DWORD   UserActiveDurationInS;
    DWORD   UserOrDispActiveDurationInS;
    DWORD   DesktopActiveDurationInS;
    WORD    Version;
    WORD    _Unk0;
    WORD    BootIdMin;
    WORD    BootIdMax;
    DWORD   PMUUKey;
    DWORD   SessionDurationInS;
    DWORD   SessionUptimeInS;
    DWORD   UserInputInS;
    DWORD   MouseInputInS;
    DWORD   KeyboardInputInS;
    DWORD   TouchInputInS;
    DWORD   PrecisionTouchpadInputInS;
    DWORD   InForegroundInS;
    DWORD   ForegroundSwitchCount;
    DWORD   UserActiveTransitionCount;
    DWORD   _Unk1;
    FILETIME    LogTimeStart;
    QWORD   CumulativeUserActiveDurationInS;
    WORD    UpdateCountAccumulationStarted;
    WORD    _Unk2;
    DWORD   BuildUserActiveDurationInS;
    DWORD   BuildNumber;
    DWORD   _UnkDeltaUserOrDispActiveDurationInS;
    DWORD   _UnkDeltaTime;
    DWORD   _Unk3;
} CIT_POST_UPDATE_USE_INFO;

// DP
typedef struct _CIT_DP_MEMOIZATION_ENTRY {
    DWORD   Unk0;
    DWORD   Unk1;
    DWORD   Unk2;
} CIT_DP_MEMOIZATION_ENTRY;

typedef struct _CIT_DP_MEMOIZATION_CONTEXT {
    _CIT_DP_MEMOIZATION_ENTRY   Entries[12];
} CIT_DP_MEMOIZATION_CONTEXT;

typedef struct _CIT_DP_DATA {
    WORD    Version;
    WORD    Size;
    WORD    LogCount;
    WORD    CrashCount;
    DWORD   SessionCount;
    DWORD   UpdateKey;
    QWORD   _Unk0;
    FILETIME    _UnkTime;
    FILETIME    LogTimeStart;
    DWORD   ForegroundDurations[11];
    DWORD   _Unk1;
    _CIT_DP_MEMOIZATION_CONTEXT MemoizationContext;
} CIT_DP_DATA;
"""

c_cit = cstruct().load(cit_def)


CITSystemRecord = TargetRecordDescriptor(
    "windows/registry/cit/system",
    [
        ("datetime", "period_start"),
        ("datetime", "start_time"),
        ("datetime", "current_time"),
        ("varint", "aggregation_period_in_s"),
        ("varint", "span_stats_context_flushes_0_count"),
        ("varint", "span_stats_context_flushes_0_duration"),
        ("varint", "span_stats_foreground_0_count"),
        ("varint", "span_stats_foreground_0_duration"),
        ("varint", "span_stats_foreground_1_count"),
        ("varint", "span_stats_foreground_1_duration"),
        ("varint", "span_stats_display_power_0_count"),
        ("varint", "span_stats_display_power_0_duration"),
        ("varint", "span_stats_display_request_change_count"),
        ("varint", "span_stats_display_request_change_duration"),
        ("varint", "span_stats_display_power_1_count"),
        ("varint", "span_stats_display_power_1_duration"),
        ("varint", "span_stats_display_power_2_count"),
        ("varint", "span_stats_display_power_2_duration"),
        ("varint", "span_stats_display_power_3_count"),
        ("varint", "span_stats_display_power_3_duration"),
        ("varint", "span_stats_context_flushes_1_count"),
        ("varint", "span_stats_context_flushes_1_duration"),
        ("varint", "span_stats_foreground_2_count"),
        ("varint", "span_stats_foreground_2_duration"),
        ("varint", "span_stats_context_flushes_2_count"),
        ("varint", "span_stats_context_flushes_2_duration"),
        ("varint", "stats_unk_boot_id_0"),
        ("varint", "stats_unk_boot_id_1"),
        ("varint", "stats_unk_boot_id_2"),
        ("varint", "stats_unk_boot_id_3"),
        ("varint", "stats_unk_boot_id_4"),
        ("varint", "stats_session_connects"),
        ("varint", "stats_process_foreground_changes"),
        ("varint", "stats_context_flushes"),
        ("varint", "stats_missing_prog_data"),
        ("varint", "stats_desktop_switches"),
        ("varint", "stats_winlogon_message"),
        ("varint", "stats_winlogon_lock_hotkey"),
        ("varint", "stats_winlogon_lock"),
        ("varint", "stats_session_disconnects"),
    ],
)


BITMAP_FIELDS = [
    ("datetime", "ts"),
    ("datetime", "period_start"),
    ("datetime", "start_time"),
    ("datetime", "current_time"),
    ("varint", "aggregation_period_in_s"),
    ("varint", "bit_period_in_s"),
]


CITSystemBitmapDisplayPowerRecord = TargetRecordDescriptor(
    "windows/registry/cit/system/bitmap/display_power",
    BITMAP_FIELDS,
)


CITSystemBitmapDisplayRequestChangeRecord = TargetRecordDescriptor(
    "windows/registry/cit/system/bitmap/display_request_change",
    BITMAP_FIELDS,
)


CITSystemBitmapInputRecord = TargetRecordDescriptor(
    "windows/registry/cit/system/bitmap/input",
    BITMAP_FIELDS,
)


CITSystemBitmapInputTouchRecord = TargetRecordDescriptor(
    "windows/registry/cit/system/bitmap/nput_touch",
    BITMAP_FIELDS,
)


CITSystemBitmapUnknownRecord = TargetRecordDescriptor(
    "windows/registry/cit/system/bitmap/unknown",
    BITMAP_FIELDS,
)


CITSystemBitmapForegroundRecord = TargetRecordDescriptor(
    "windows/registry/cit/system/bitmap/foreground",
    BITMAP_FIELDS,
)

CITProgramRecord = TargetRecordDescriptor(
    "windows/registry/cit/program",
    [
        ("datetime", "period_start"),
        ("datetime", "start_time"),
        ("datetime", "current_time"),
        ("varint", "aggregation_period_in_s"),
        ("path", "path"),
        ("string", "command_line"),
        ("datetime", "pe_timedatestamp"),
        ("varint", "pe_checksum"),
        ("varint", "extra3"),
        ("varint", "span_stats_process_creation_0_count"),
        ("varint", "span_stats_process_creation_0_duration"),
        ("varint", "span_stats_foreground_0_count"),
        ("varint", "span_stats_foreground_0_duration"),
        ("varint", "span_stats_foreground_1_count"),
        ("varint", "span_stats_foreground_1_duration"),
        ("varint", "span_stats_foreground_2_count"),
        ("varint", "span_stats_foreground_2_duration"),
        ("varint", "span_stats_process_suspended_count"),
        ("varint", "span_stats_process_suspended_duration"),
        ("varint", "span_stats_process_creation_1_count"),
        ("varint", "span_stats_process_creation_1_duration"),
        ("varint", "stats_crashes"),
        ("varint", "stats_thread_ghosting_changes"),
        ("varint", "stats_input"),
        ("varint", "stats_input_keyboard"),
        ("varint", "stats_unknown"),
        ("varint", "stats_input_touch"),
        ("varint", "stats_input_hid"),
        ("varint", "stats_input_mouse"),
        ("varint", "stats_mouse_left_button"),
        ("varint", "stats_mouse_right_button"),
        ("varint", "stats_mouse_middle_button"),
        ("varint", "stats_mouse_wheel"),
    ],
)


CITProgramBitmapForegroundRecord = TargetRecordDescriptor(
    "windows/registry/cit/program/bitmap/foreground",
    BITMAP_FIELDS,
)


CITRecords = Union[  # noqa: UP007
    CITSystemRecord,
    CITSystemBitmapDisplayPowerRecord,
    CITSystemBitmapDisplayRequestChangeRecord,
    CITSystemBitmapInputRecord,
    CITSystemBitmapInputTouchRecord,
    CITSystemBitmapUnknownRecord,
    CITSystemBitmapForegroundRecord,
    CITProgramRecord,
]


CITPostUpdateUseInfoRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "windows/registry/cit/puu",
    [
        ("datetime", "log_time_start"),
        ("varint", "update_key"),
        ("varint", "update_count"),
        ("varint", "crash_count"),
        ("varint", "session_count"),
        ("varint", "log_count"),
        ("varint", "user_active_duration_in_s"),
        ("varint", "user_or_display_active_duration_in_s"),
        ("varint", "desktop_active_duration_in_s"),
        ("varint", "puu_version"),
        ("varint", "boot_id_min"),
        ("varint", "boot_id_max"),
        ("varint", "pmuu_key"),
        ("varint", "session_duration_in_s"),
        ("varint", "session_uptime_in_s"),
        ("varint", "user_input_in_s"),
        ("varint", "mouse_input_in_s"),
        ("varint", "keyboard_input_in_s"),
        ("varint", "touch_input_in_s"),
        ("varint", "precision_touchpad_input_in_s"),
        ("varint", "in_foreground_in_s"),
        ("varint", "foreground_switch_count"),
        ("varint", "user_active_transition_count"),
        ("varint", "cumulative_user_active_duration_in_s"),
        ("varint", "update_count_accumulation_started"),
        ("varint", "build_user_active_duration_in_s"),
        ("varint", "build_number"),
    ],
)


CITDPRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "windows/registry/cit/dp",
    [
        ("datetime", "log_time_start"),
        ("varint", "update_key"),
        ("varint", "log_count"),
        ("varint", "crash_count"),
        ("varint", "session_count"),
    ],
)


CITDPDurationRecord = create_extended_descriptor([UserRecordDescriptorExtension])(
    "windows/registry/cit/dp/duration",
    [
        ("datetime", "log_time_start"),
        ("string", "application"),
        ("varint", "duration"),
    ],
)


CITTelemetryRecord = TargetRecordDescriptor(
    "windows/registry/cit/telemetry",
    [
        ("datetime", "regf_mtime"),
        ("varint", "telemetry_version"),
        ("path", "path"),
        ("string", "value"),
    ],
)


CITModuleRecord = TargetRecordDescriptor(
    "windows/registry/cit/module",
    [
        ("datetime", "last_loaded"),
        ("datetime", "regf_mtime"),
        ("path", "tracked_module"),
        ("path", "executable"),
        ("datetime", "overflow_quota"),
        ("datetime", "overflow_value"),
    ],
)


class CIT:
    def __init__(self, buf: bytes):
        compressed_fh = BytesIO(buf)
        # Compressed size, uncompressed size
        _, _ = struct.unpack("<2I", compressed_fh.read(8))

        self.buf = lznt1.decompress(compressed_fh)

        self.header = c_cit.CIT_HEADER(self.buf)
        if self.header.MajorVersion != 0x0A:
            raise ValueError("Unsupported CIT version")

        digest = crc32(self.buf[0x14:], crc32(self.buf[:0x10]))
        if self.header.Crc32 != digest:
            raise ValueError("Crc32 mismatch")

        system_data_buf = self.data(self.header.SystemDataOffset, self.header.SystemDataSize, 0x18)
        self.system_data = SystemData(self, c_cit.CIT_PERSISTED(system_data_buf))

        base_use_data_buf = self.data(self.header.BaseUseDataOffset, self.header.BaseUseDataSize, 0x18)
        self.base_use_data = BaseUseData(self, c_cit.CIT_PERSISTED(base_use_data_buf))

        entry_data = self.buf[self.header.EntryDataOffset :]
        self.entries = [Entry(self, entry) for entry in c_cit.CIT_ENTRY[self.header.EntryCount](entry_data)]

    def data(self, offset: int, size: int, expected_size: int | None = None) -> bytes:
        if expected_size and size > expected_size:
            size = expected_size

        data = self.buf[offset : offset + size]

        if expected_size and size < expected_size:
            data.ljust(expected_size, b"\x00")

        return data

    def iter_bitmap(self, bitmap: bytes) -> Iterator[datetime.datetime]:
        bit_delta = datetime.timedelta(seconds=self.header.BitPeriodInS)
        ts = wintimestamp(self.header.PeriodStartLocal).replace(tzinfo=None)

        for byte in bitmap:
            if byte == b"\x00":
                ts += 8 * bit_delta
            else:
                for bit in range(8):
                    if byte & (1 << bit):
                        yield ts
                    ts += bit_delta


class Entry:
    def __init__(self, cit: CIT, entry: c_cit.CIT_ENTRY):
        self.cit = cit
        self.entry = entry

        program_buf = cit.data(entry.ProgramDataOffset, entry.ProgramDataSize, 0x1C)
        self.program_data = c_cit.CIT_PROGRAM_DATA(program_buf)

        use_data_buf = cit.data(entry.UseDataOffset, entry.UseDataSize, 0x18)
        self.use_data = ProgramUseData(cit, c_cit.CIT_PERSISTED(use_data_buf))

        self.file_path = None
        self.command_line = None

        if self.program_data.FilePathOffset:
            file_path_buf = cit.data(self.program_data.FilePathOffset, self.program_data.FilePathSize * 2)
            self.file_path = file_path_buf.decode("utf-16-le")

        if self.program_data.CommandLineOffset:
            command_line_buf = cit.data(self.program_data.CommandLineOffset, self.program_data.CommandLineSize * 2)
            self.command_line = command_line_buf.decode("utf-16-le")

    def __repr__(self) -> str:
        return f"<Entry file_path={self.file_path!r} command_line={self.command_line!r}>"


class BaseUseData:
    MIN_BITMAPS_SIZE = 0x8
    MIN_SPAN_STATS_SIZE = 0x30
    MIN_STATS_SIZE = 0x18

    def __init__(self, cit: CIT, entry: c_cit.CIT_ENTRY):
        self.cit = cit
        self.entry = entry

        bitmap_items = c_cit.CIT_BITMAP_ITEM[entry.BitmapsSize // len(c_cit.CIT_BITMAP_ITEM)](
            cit.data(entry.BitmapsOffset, entry.BitmapsSize, self.MIN_BITMAPS_SIZE)
        )
        bitmaps = [cit.data(item.Offset, item.Size) for item in bitmap_items]
        self.bitmaps = self._parse_bitmaps(bitmaps)
        self.span_stats = self._parse_span_stats(
            cit.data(entry.SpanStatsOffset, entry.SpanStatsSize, self.MIN_SPAN_STATS_SIZE)
        )
        self.stats = self._parse_stats(cit.data(entry.StatsOffset, entry.StatsSize, self.MIN_STATS_SIZE))

    def _parse_bitmaps(self, bitmaps: list[bytes]) -> BaseUseDataBitmaps:
        return BaseUseDataBitmaps(self.cit, bitmaps)

    def _parse_span_stats(self, span_stats_data: bytes) -> None:
        return None

    def _parse_stats(self, stats_data: bytes) -> None:
        return None


class BaseUseDataBitmaps:
    def __init__(self, cit: CIT, bitmaps: list[bytes]):
        self.cit = cit
        self._bitmaps = bitmaps

    def _parse_bitmap(self, idx: int) -> list[datetime.datetime]:
        return list(self.cit.iter_bitmap(self._bitmaps[idx]))


class SystemData(BaseUseData):
    MIN_BITMAPS_SIZE = 0x30
    MIN_SPAN_STATS_SIZE = 0x58
    MIN_STATS_SIZE = 0x1C

    def _parse_bitmaps(self, bitmaps: list[bytes]) -> SystemDataBitmaps:
        return SystemDataBitmaps(self.cit, bitmaps)

    def _parse_span_stats(self, span_stats_data: bytes) -> c_cit.CIT_SYSTEM_DATA_SPAN_STATS:
        return c_cit.CIT_SYSTEM_DATA_SPAN_STATS(span_stats_data)

    def _parse_stats(self, stats_data: bytes) -> c_cit.CIT_SYSTEM_DATA_STATS:
        return c_cit.CIT_SYSTEM_DATA_STATS(stats_data)


class SystemDataBitmaps(BaseUseDataBitmaps):
    def __init__(self, cit: CIT, bitmaps: list[bytes]):
        super().__init__(cit, bitmaps)
        self.display_power = self._parse_bitmap(0)
        self.display_request_change = self._parse_bitmap(1)
        self.input = self._parse_bitmap(2)
        self.input_touch = self._parse_bitmap(3)
        self.unknown = self._parse_bitmap(4)
        self.foreground = self._parse_bitmap(5)


class ProgramUseData(BaseUseData):
    def _parse_bitmaps(self, bitmaps: list[bytes]) -> ProgramDataBitmaps:
        return ProgramDataBitmaps(self.cit, bitmaps)

    def _parse_span_stats(self, span_stats_data: bytes) -> c_cit.CIT_USE_DATA_SPAN_STATS:
        return c_cit.CIT_USE_DATA_SPAN_STATS(span_stats_data)

    def _parse_stats(self, stats_data: bytes) -> c_cit.CIT_USE_DATA_STATS:
        return c_cit.CIT_USE_DATA_STATS(stats_data)


class ProgramDataBitmaps(BaseUseDataBitmaps):
    def __init__(self, cit: CIT, use_data: list[bytes]):
        super().__init__(cit, use_data)
        self.foreground = self._parse_bitmap(0)


def decode_name(name: str) -> bytes:
    """Decode the registry key name.

    The CIT key name in the registry has some strange encoding.
    This function is currently unused, but leaving it here for reference if someone needs it.
    """
    buf = name.encode()
    out = []

    for idx in range(0, len(name), 2):
        c0 = buf[idx]
        c1 = buf[idx + 1]

        c0 -= 0x30 if c0 <= 0x39 else 0x37
        c1 -= 0x30 if c1 <= 0x39 else 0x37

        out.append(c0 | (c1 * 16))

    return bytes(out)


def local_wintimestamp(target: Target, ts: int) -> datetime.datetime:
    return target.datetime.to_utc(wintimestamp(ts))


class CITPlugin(Plugin):
    """Plugin that parses CIT data from the registry.

    References:
        - https://dfir.ru/2018/12/02/the-cit-database-and-the-syscache-hive/
    """

    __namespace__ = "cit"

    KEY = "HKLM\\Software\\Microsoft\\Windows NT\\CurrentVersion\\AppCompatFlags\\CIT"

    def check_compatible(self) -> None:
        if not list(self.target.registry.keys(self.KEY)):
            raise UnsupportedPluginError("No CIT registry key found")

    @export(record=get_args(CITRecords))
    def cit(self) -> Iterator[CITRecords]:
        """Return CIT data from the registry for executed executable information.

        CIT data is stored at HKLM\\Software\\Microsoft\\Windows NT\\CurrentVersion\\AppCompatFlags\\CIT\\System.
        It's supposedly application usage data that has yet-to-be flushed to the amcache.

        Some of its values are still unknown. Generally only available before Windows 10.
        """

        for key in self.target.registry.keys(f"{self.KEY}\\System"):
            for value in key.values():
                # The value name is a weird per-byte character-mirrored value that contains copies
                # of the PeriodStartLocal and AggregationperiodInS fields
                # The CIT\\Proc subkey also sometimes has a timestamp in it, which appears to be the time
                # it was processed by generaltel.dll

                data = value.value
                if len(data) <= 8:
                    continue

                try:
                    cit = CIT(data)

                    sysdata = cit.system_data
                    yield CITSystemRecord(
                        period_start=local_wintimestamp(self.target, cit.header.PeriodStartLocal),
                        start_time=local_wintimestamp(self.target, cit.header.StartTimeLocal),
                        current_time=local_wintimestamp(self.target, cit.header.CurrentTimeLocal),
                        aggregation_period_in_s=cit.header.AggregationPeriodInS,
                        span_stats_context_flushes_0_count=sysdata.span_stats.ContextFlushes0.Count,
                        span_stats_context_flushes_0_duration=sysdata.span_stats.ContextFlushes0.Duration,
                        span_stats_foreground_0_count=sysdata.span_stats.Foreground0.Count,
                        span_stats_foreground_0_duration=sysdata.span_stats.Foreground0.Duration,
                        span_stats_foreground_1_count=sysdata.span_stats.Foreground1.Count,
                        span_stats_foreground_1_duration=sysdata.span_stats.Foreground1.Duration,
                        span_stats_display_power_0_count=sysdata.span_stats.DisplayPower0.Count,
                        span_stats_display_power_0_duration=sysdata.span_stats.DisplayPower0.Duration,
                        span_stats_display_request_change_count=sysdata.span_stats.DisplayRequestChange.Count,
                        span_stats_display_request_change_duration=sysdata.span_stats.DisplayRequestChange.Duration,
                        span_stats_display_power_1_count=sysdata.span_stats.DisplayPower1.Count,
                        span_stats_display_power_1_duration=sysdata.span_stats.DisplayPower2.Duration,
                        span_stats_display_power_2_count=sysdata.span_stats.DisplayPower2.Count,
                        span_stats_display_power_2_duration=sysdata.span_stats.DisplayPower2.Duration,
                        span_stats_display_power_3_count=sysdata.span_stats.DisplayPower3.Count,
                        span_stats_display_power_3_duration=sysdata.span_stats.DisplayPower3.Duration,
                        span_stats_context_flushes_1_count=sysdata.span_stats.ContextFlushes1.Count,
                        span_stats_context_flushes_1_duration=sysdata.span_stats.ContextFlushes1.Duration,
                        span_stats_foreground_2_count=sysdata.span_stats.Foreground2.Count,
                        span_stats_foreground_2_duration=sysdata.span_stats.Foreground2.Duration,
                        span_stats_context_flushes_2_count=sysdata.span_stats.ContextFlushes2.Count,
                        span_stats_context_flushes_2_duration=sysdata.span_stats.ContextFlushes2.Duration,
                        stats_unk_boot_id_0=sysdata.stats.Unknown_BootIdRelated0,
                        stats_unk_boot_id_1=sysdata.stats.Unknown_BootIdRelated1,
                        stats_unk_boot_id_2=sysdata.stats.Unknown_BootIdRelated2,
                        stats_unk_boot_id_3=sysdata.stats.Unknown_BootIdRelated3,
                        stats_unk_boot_id_4=sysdata.stats.Unknown_BootIdRelated4,
                        stats_session_connects=sysdata.stats.SessionConnects,
                        stats_process_foreground_changes=sysdata.stats.ProcessForegroundChanges,
                        stats_context_flushes=sysdata.stats.ContextFlushes,
                        stats_missing_prog_data=sysdata.stats.MissingProgData,
                        stats_desktop_switches=sysdata.stats.DesktopSwitches,
                        stats_winlogon_message=sysdata.stats.WinlogonMessage,
                        stats_winlogon_lock_hotkey=sysdata.stats.WinlogonLockHotkey,
                        stats_winlogon_lock=sysdata.stats.WinlogonLock,
                        stats_session_disconnects=sysdata.stats.SessionDisconnects,
                        _target=self.target,
                    )

                    bitmap_pairs = [
                        (sysdata.bitmaps.display_power, CITSystemBitmapDisplayPowerRecord),
                        (sysdata.bitmaps.display_request_change, CITSystemBitmapDisplayRequestChangeRecord),
                        (sysdata.bitmaps.input, CITSystemBitmapInputRecord),
                        (sysdata.bitmaps.input_touch, CITSystemBitmapInputTouchRecord),
                        (sysdata.bitmaps.unknown, CITSystemBitmapUnknownRecord),
                        (sysdata.bitmaps.foreground, CITSystemBitmapForegroundRecord),
                    ]

                    for bitmap, record in bitmap_pairs:
                        yield from _yield_bitmap_records(self.target, cit, bitmap, record)

                    for entry in cit.entries:
                        program_data = entry.program_data
                        span_stats = entry.use_data.span_stats
                        stats = entry.use_data.stats

                        yield CITProgramRecord(
                            period_start=local_wintimestamp(self.target, cit.header.PeriodStartLocal),
                            start_time=local_wintimestamp(self.target, cit.header.StartTimeLocal),
                            current_time=local_wintimestamp(self.target, cit.header.CurrentTimeLocal),
                            aggregation_period_in_s=cit.header.AggregationPeriodInS,
                            path=self.target.fs.path(entry.file_path),
                            command_line=entry.command_line,
                            pe_timedatestamp=program_data.PeTimeDateStamp,
                            pe_checksum=program_data.PeCheckSum,
                            extra3=program_data.Extra3,
                            span_stats_process_creation_0_count=span_stats.ProcessCreation0.Count,
                            span_stats_process_creation_0_duration=span_stats.ProcessCreation0.Duration,
                            span_stats_foreground_0_count=span_stats.Foreground0.Count,
                            span_stats_foreground_0_duration=span_stats.Foreground0.Duration,
                            span_stats_foreground_1_count=span_stats.Foreground1.Count,
                            span_stats_foreground_1_duration=span_stats.Foreground1.Duration,
                            span_stats_foreground_2_count=span_stats.Foreground2.Count,
                            span_stats_foreground_2_duration=span_stats.Foreground2.Duration,
                            span_stats_process_suspended_count=span_stats.ProcessSuspended.Count,
                            span_stats_process_suspended_duration=span_stats.ProcessSuspended.Duration,
                            span_stats_process_creation_1_count=span_stats.ProcessCreation1.Count,
                            span_stats_process_creation_1_duration=span_stats.ProcessCreation1.Duration,
                            stats_crashes=stats.Crashes,
                            stats_thread_ghosting_changes=stats.ThreadGhostingChanges,
                            stats_input=stats.Input,
                            stats_input_keyboard=stats.InputKeyboard,
                            stats_unknown=stats.Unknown,
                            stats_input_touch=stats.InputTouch,
                            stats_input_hid=stats.InputHid,
                            stats_input_mouse=stats.InputMouse,
                            stats_mouse_left_button=stats.MouseLeftButton,
                            stats_mouse_right_button=stats.MouseRightButton,
                            stats_mouse_middle_button=stats.MouseMiddleButton,
                            stats_mouse_wheel=stats.MouseWheel,
                            _target=self.target,
                        )

                        yield from _yield_bitmap_records(
                            self.target, cit, entry.use_data.bitmaps.foreground, CITProgramBitmapForegroundRecord
                        )
                except Exception as e:
                    self.target.log.warning("Failed to parse CIT value: %s", value.name)
                    self.target.log.debug("", exc_info=e)

    @export(record=CITPostUpdateUseInfoRecord)
    def puu(self) -> Iterator[CITPostUpdateUseInfoRecord]:
        """Parse CIT PUU (Post Update Usage) data from the registry.

        Generally only available since Windows 10.
        """

        keys = [
            self.KEY,
            "HKCU\\Software\\Microsoft\\Windows NT\\CurrentVersion\\Winlogon",
        ]

        for reg_key in keys:
            for key in self.target.registry.keys(reg_key):
                try:
                    key_value = key.value("PUUActive").value
                    puu = c_cit.CIT_POST_UPDATE_USE_INFO(key_value)
                except RegistryValueNotFoundError:
                    continue

                except EOFError as e:
                    self.target.log.warning("Exception reading CIT structure in key %s", key.path)
                    self.target.log.debug("Unable to parse value %s", key_value, exc_info=e)
                    continue

                yield CITPostUpdateUseInfoRecord(
                    log_time_start=wintimestamp(puu.LogTimeStart),
                    update_key=puu.UpdateKey,
                    update_count=puu.UpdateCount,
                    crash_count=puu.CrashCount,
                    session_count=puu.SessionCount,
                    log_count=puu.LogCount,
                    user_active_duration_in_s=puu.UserActiveDurationInS,
                    user_or_display_active_duration_in_s=puu.UserOrDispActiveDurationInS,
                    desktop_active_duration_in_s=puu.DesktopActiveDurationInS,
                    puu_version=puu.Version,
                    boot_id_min=puu.BootIdMin,
                    boot_id_max=puu.BootIdMax,
                    pmuu_key=puu.PMUUKey,
                    session_duration_in_s=puu.SessionDurationInS,
                    session_uptime_in_s=puu.SessionUptimeInS,
                    user_input_in_s=puu.UserInputInS,
                    mouse_input_in_s=puu.MouseInputInS,
                    keyboard_input_in_s=puu.KeyboardInputInS,
                    touch_input_in_s=puu.TouchInputInS,
                    precision_touchpad_input_in_s=puu.PrecisionTouchpadInputInS,
                    in_foreground_in_s=puu.InForegroundInS,
                    foreground_switch_count=puu.ForegroundSwitchCount,
                    user_active_transition_count=puu.UserActiveTransitionCount,
                    cumulative_user_active_duration_in_s=puu.CumulativeUserActiveDurationInS,
                    update_count_accumulation_started=puu.UpdateCountAccumulationStarted,
                    build_user_active_duration_in_s=puu.BuildUserActiveDurationInS,
                    build_number=puu.BuildNumber,
                    _target=self.target,
                    _user=self.target.registry.get_user(key),
                )

    @export(record=[CITDPRecord, CITDPDurationRecord])
    def dp(self) -> Iterator[CITDPRecord | CITDPDurationRecord]:
        """Parse CIT DP data from the registry.

        Generally only available since Windows 10.
        """

        keys = [
            self.KEY,
            "HKCU\\Software\\Microsoft\\Windows NT\\CurrentVersion\\Winlogon",
        ]
        applications = [
            ["Cummulative"],
            ["IEXPLORE.EXE"],
            ["MICROSOFTEDGE.EXE", "MICROSOFTEDGECP.EXE", "MICROSOFTEDGEBCHOST.EXE", "MICROSOFTEDGEDEVTOOLS.EXE"],
            ["CHROME.EXE"],
            ["WINWORD.EXE"],
            ["EXCEL.EXE"],
            ["FIREFOX.EXE"],
            ["MICROSOFT.PHOTOS.EXE"],
            ["OUTLOOK.EXE"],
            ["ACRORD32.EXE"],
            ["SKYPE.EXE"],
        ]

        for reg_key in keys:
            for key in self.target.registry.keys(reg_key):
                try:
                    key_value = key.value("DP").value
                    dp = c_cit.CIT_DP_DATA(key_value)
                except RegistryValueNotFoundError:
                    continue

                except EOFError as e:
                    self.target.log.warning("Exception reading CIT structure in key %s", key.path)
                    self.target.log.debug("Unable to parse value %s", key_value, exc_info=e)
                    continue

                user = self.target.registry.get_user(key)
                log_time_start = wintimestamp(dp.LogTimeStart)

                yield CITDPRecord(
                    log_time_start=log_time_start,
                    update_key=dp.UpdateKey,
                    log_count=dp.LogCount,
                    crash_count=dp.CrashCount,
                    session_count=dp.SessionCount,
                    _target=self.target,
                    _user=user,
                )

                for names, duration in zip(applications, dp.ForegroundDurations, strict=False):
                    for name in names:
                        yield CITDPDurationRecord(
                            log_time_start=log_time_start,
                            application=name,
                            duration=duration,
                            _target=self.target,
                            _user=user,
                        )

    @export(record=CITTelemetryRecord)
    def telemetry(self) -> Iterator[CITTelemetryRecord]:
        """Parse CIT process telemetry answers from the registry.

        In some versions of Windows, processes would get "telemetry answers" set on their process struct, based on
        if certain events happened.

        Generally only available before Windows 10.
        """

        for key in self.target.registry.keys(f"{self.KEY}\\win32k"):
            for version_key in key.subkeys():
                for value in version_key.values():
                    yield CITTelemetryRecord(
                        regf_mtime=version_key.ts,
                        telemetry_version=version_key.name,
                        path=self.target.fs.path(value.name),
                        value=str(c_cit.TELEMETRY_ANSWERS(value.value)).split(".")[1],
                        _target=self.target,
                    )

    @export(record=CITModuleRecord)
    def modules(self) -> Iterator[CITModuleRecord]:
        """Parse CIT tracked module information from the registry.

        Contains applications that loaded a tracked module. By default these are:

            \\System32\\mrt100.dll
            Microsoft.NET\\Framework\\v1.0.3705\\mscorwks.dll
            Microsoft.NET\\Framework\\v1.0.3705\\mscorsvr.dll
            Microsoft.NET\\Framework\\v1.1.4322\\mscorwks.dll
            Microsoft.NET\\Framework\\v1.1.4322\\mscorsvr.dll
            Microsoft.NET\\Framework\\v2.0.50727\\mscorwks.dll
            \\Microsoft.NET\\Framework\\v4.0.30319\\clr.dll
            \\Microsoft.NET\\Framework64\\v4.0.30319\\clr.dll
            \\Microsoft.NET\\Framework64\\v2.0.50727\\mscorwks.dll

        When the amount of executables exceeds 64, the OverflowQuota value is set with the last timestamp.
        When the path length of an executable exceeds 520 characters, the OverflowValue value is set.

        Generally only available since Windows 10.
        """

        for key in self.target.registry.keys(f"{self.KEY}\\Module"):
            for monitored_dll in key.subkeys():
                try:
                    overflow_quota = wintimestamp(monitored_dll.value("OverflowQuota").value)
                except RegistryValueNotFoundError:
                    overflow_quota = None

                try:
                    overflow_value = wintimestamp(monitored_dll.value("OverflowValue").value)
                except RegistryValueNotFoundError:
                    overflow_value = None

                for value in monitored_dll.values():
                    if value.name in ("OverflowQuota", "OverflowValue"):
                        continue

                    yield CITModuleRecord(
                        last_loaded=wintimestamp(value.value),
                        regf_mtime=monitored_dll.ts,
                        tracked_module=self.target.fs.path(monitored_dll.name),
                        executable=self.target.fs.path(value.name),
                        # These are actually specific for the tracked module, but just include them in every record
                        overflow_quota=overflow_quota,
                        overflow_value=overflow_value,
                        _target=self.target,
                    )


def _yield_bitmap_records(
    target: Target, cit: CIT, bitmap: list[datetime.datetime], record: TargetRecordDescriptor
) -> Iterator[Record]:
    for entry in bitmap:
        yield record(
            ts=target.datetime.to_utc(entry),
            period_start=local_wintimestamp(target, cit.header.PeriodStartLocal),
            start_time=local_wintimestamp(target, cit.header.StartTimeLocal),
            current_time=local_wintimestamp(target, cit.header.CurrentTimeLocal),
            aggregation_period_in_s=cit.header.AggregationPeriodInS,
            bit_period_in_s=cit.header.BitPeriodInS,
            _target=target,
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.localeutil import normalize_language, normalize_timezone
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import export
from dissect.target.plugins.os.default.locale import LocalePlugin

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

WindowsKeyboardRecord = TargetRecordDescriptor(
    "windows/keyboard",
    [
        ("string", "layout"),
        ("string", "language_id"),
    ],
)


class WindowsLocalePlugin(LocalePlugin):
    """Windows locale plugin."""

    def __init__(self, target: Target):
        super().__init__(target)
        self.LANG_DICT = {
            lang.name: lang.value
            for k in self.target.registry.key("HKLM\\SYSTEM\\CurrentControlSet\\Control\\Keyboard Layout\\DosKeybCodes")
            for lang in k.values()
        }

    def check_compatible(self) -> None:
        if not self.target.has_function("registry"):
            raise UnsupportedPluginError("Unsupported Plugin")

    @export(property=True)
    def timezone(self) -> str | None:
        """Get the configured timezone of the system in IANA TZ standard format."""
        return normalize_timezone(self.target.datetime.tzinfo.name)

    @export(property=True)
    def language(self) -> str | None:
        """Get a list of installed languages on the system."""
        # HKCU\\Control Panel\\International\\User Profile" Languages
        found_languages = []
        for up in self.target.registry.keys("HKCU\\Control Panel\\International\\User Profile"):
            for subkey in up.subkeys():
                language = normalize_language(subkey.name.replace("-", "_"))
                if language not in found_languages:
                    found_languages.append(language)
        return found_languages

    @export(record=WindowsKeyboardRecord)
    def keyboard(self) -> Iterator[WindowsKeyboardRecord]:
        """Yield records of installed keyboards on the system."""
        found_keyboards = []
        for key in self.target.registry.keys("HKCU\\Keyboard Layout\\Preload"):
            for language in key.values():
                language_id = language.value
                if language_id not in found_keyboards:
                    found_keyboards.append(language_id)
                    yield WindowsKeyboardRecord(
                        layout=self.LANG_DICT.get(language_id, language_id),
                        language_id=language_id,
                        _target=self.target,
                    )



from __future__ import annotations

import re
from typing import TYPE_CHECKING

from dissect.util.ts import from_unix

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.fsutil import basename, open_decompress
from dissect.target.helpers.record import TargetRecordDescriptor
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from dissect.target.target import Target

AuditRecord = TargetRecordDescriptor(
    "linux/log/audit",
    [
        ("datetime", "ts"),
        ("string", "audit_type"),
        ("varint", "audit_id"),
        ("string", "message"),
        ("path", "source"),
    ],
)

AUDIT_REGEX = re.compile(r"^type=(?P<audit_type>.*) msg=audit\((?P<ts>.*):(?P<audit_id>.*)\): (?P<message>.*)$")


class AuditPlugin(Plugin):
    """Unix audit log plugin."""

    def __init__(self, target: Target):
        super().__init__(target)
        self.log_paths = self.get_log_paths()

    def check_compatible(self) -> None:
        if not len(self.log_paths):
            raise UnsupportedPluginError("No audit path found")

    def get_log_paths(self) -> list[Path]:
        log_paths = []

        log_paths.extend(self.target.fs.path("/var/log/audit").glob("audit.log*"))

        default_config = "/etc/audit/auditd.conf"
        if (path := self.target.fs.path(default_config)).exists():
            for line in path.open("rt"):
                line = line.strip()
                if not line or "log_file" not in line:
                    continue

                log_path = line.split("=")[-1].strip()
                parent_folder = self.target.fs.path(log_path).parent
                log_paths.extend(path for path in parent_folder.glob(f"{basename(log_path)}*") if path not in log_paths)

        return log_paths

    @export(record=AuditRecord)
    def audit(self) -> Iterator[AuditRecord]:
        """Return CentOS and RedHat audit information stored in /var/log/audit*.

        The audit log file on a Linux machine stores security-relevant information.
        Based on pre-configured rules. Log messages consist of space delimited key=value pairs.

        References:
            - https://access.redhat.com/documentation/en-us/red_hat_enterprise_linux/6/html/security_guide/chap-system_auditing
            - https://linux-audit.com/linux-audit-log-files-in-var-log-audit/
            - https://man7.org/linux/man-pages/man8/auditd.8.html
            - https://man7.org/linux/man-pages/man8/ausearch.8.html
            - https://access.redhat.com/documentation/en-us/red_hat_enterprise_linux/7/html/security_guide/sec-understanding_audit_log_files
        """

        for path in self.log_paths:
            try:
                path = path.resolve(strict=True)
                for line in open_decompress(path, "rt"):
                    match = AUDIT_REGEX.match(line)
                    if not match:
                        self.target.log.warning("Audit log file contains unrecognized format in %s", path)
                        continue

                    match = match.groupdict()
                    yield AuditRecord(
                        ts=from_unix(float(match["ts"])),
                        audit_type=match["audit_type"],
                        audit_id=int(match["audit_id"]),
                        message=match["message"].strip(),
                        source=path,
                        _target=self.target,
                    )
            except FileNotFoundError:
                self.target.log.warning("Audit log file configured but could not be found (dead symlink?): %s", path)
            except Exception as e:
                self.target.log.warning("An error occured parsing audit log file %s: %s", path, str(e))
                self.target.log.debug("", exc_info=e)



from __future__ import annotations

import io

from dissect.target import container
from dissect.target.containers.ewf import EwfContainer
from dissect.target.filesystem import VirtualFilesystem
from tests._utils import absolute_path


def test_ewf_container() -> None:
    """Test that EWF containers are properly opened.

    Generated with::

        echo "testdissecte01" | ewfacquirestream -t small
    """
    path = absolute_path("_data/containers/ewf/small.E01")

    fh = container.open(path)
    assert isinstance(fh, EwfContainer)
    a = fh.read(20)
    assert a == b"testdissecte01\n"
    assert fh.tell() == 15
    fh.seek(0, whence=io.SEEK_SET)
    assert fh.read(20) == b"testdissecte01\n"
    fh.close()


def test_ewf_detect_path() -> None:
    """Test that EWF containers are properly opened, when using the extension based matching."""
    vfs = VirtualFilesystem()
    vfs.map_file("small.E01", absolute_path("_data/containers/ewf/small.E01"))
    fh = container.open(vfs.path("small.E01"))
    assert isinstance(fh, EwfContainer)
    a = fh.read(20)
    assert a == b"testdissecte01\n"
    assert fh.tell() == 15
    fh.seek(0, whence=io.SEEK_SET)
    assert fh.read(20) == b"testdissecte01\n"


def test_ewf_container_splitted() -> None:
    """Test that EWF containers are properly opened when container is split among multiple files.

    Generated with::

        # ewfacquire has a minimum 1mb split size, we generate a bit more than 1Mb of compressed data
        dd if=/dev/urandom bs=512 count=2100 of=random
        cat random | ewfacquirestream -t split -S 1048576
    """
    path = absolute_path("_data/containers/ewf/split.E01")

    fh = container.open(path)
    assert isinstance(fh, EwfContainer)
    assert fh.read(20) == b'T\x0e\xf4x\x9eT\x17\xda\xca\xbfV\xbb\xda\x99"\xe7S\xa8J\xe7'
    fh.seek(0, whence=io.SEEK_END)
    assert fh.tell() == 1075200
    fh.seek(-15, whence=io.SEEK_CUR)
    assert fh.read(20) == b"L5\x089\x8c\xe4\x91~\x9a4\xbcG@\xb4\x11"
    fh.close()


def test_ewf_split_detect_path() -> None:
    """Test that EWF containers are properly opened, when using the extension based matching on multiple files."""
    vfs = VirtualFilesystem()
    vfs.map_file("split.E01", absolute_path("_data/containers/ewf/split.E01"))
    vfs.map_file("split.E02", absolute_path("_data/containers/ewf/split.E02"))
    fh = container.open(vfs.path("split.E01"))
    assert isinstance(fh, EwfContainer)
    assert fh.read(20) == b'T\x0e\xf4x\x9eT\x17\xda\xca\xbfV\xbb\xda\x99"\xe7S\xa8J\xe7'
    fh.seek(0, whence=io.SEEK_END)
    assert fh.tell() == 1075200
    fh.seek(-15, whence=io.SEEK_CUR)
    assert fh.read(20) == b"L5\x089\x8c\xe4\x91~\x9a4\xbcG@\xb4\x11"



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import (
    COMMON_APPLICATION_FIELDS,
    TargetRecordDescriptor,
)
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from dissect.target.target import Target

WindowsApplicationRecord = TargetRecordDescriptor(
    "windows/application",
    COMMON_APPLICATION_FIELDS,
)


class WindowsApplicationsPlugin(Plugin):
    """Windows Applications plugin."""

    def __init__(self, target: Target):
        super().__init__(target)
        self.keys = list(self.target.registry.keys("HKLM\\SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Uninstall"))

    def check_compatible(self) -> None:
        if not self.target.has_function("registry"):
            raise UnsupportedPluginError("No Windows registry found")

        if not self.keys:
            raise UnsupportedPluginError("No 'Uninstall' registry keys found")

    @export(record=WindowsApplicationRecord)
    def applications(self) -> Iterator[WindowsApplicationRecord]:
        """Yields currently installed applications from the Windows registry.

        Use the Windows eventlog plugin (``evtx``, ``evt``) to parse install and uninstall events
        of applications and services (e.g. ``4697``, ``110707``, ``1034`` and ``11724``).

        References:
            - https://learn.microsoft.com/en-us/windows/win32/msi/uninstall-registry-key

        Yields ``WindowsApplicationRecord`` records with the following fields:

        .. code-block:: text

            ts_modified  (datetime): timestamp when the installation was modified according to the registry
            ts_installed (datetime): the date when the application was installed according to the application
            name         (string):   name of the application
            version      (string):   version of the application
            author       (string):   author of the application
            type         (string):   type of the application, either user or system
            path         (string):   path to the installed location or installer of the application
        """
        for uninstall in self.keys:
            for app in uninstall.subkeys():
                values = {value.name: value.value for value in app.values()}

                if install_date := values.get("InstallDate"):
                    install_date = datetime.datetime.strptime(install_date, "%Y%m%d").replace(
                        tzinfo=datetime.timezone.utc
                    )

                yield WindowsApplicationRecord(
                    ts_modified=app.ts,
                    ts_installed=install_date,
                    name=values.get("DisplayName") or app.name,
                    version=values.get("DisplayVersion"),
                    author=values.get("Publisher"),
                    type="system" if values.get("SystemComponent") or not values else "user",
                    path=values.get("DisplayIcon") or values.get("InstallLocation") or values.get("InstallSource"),
                    _target=self.target,
                )



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.target.helpers.record import EmptyRecord
from dissect.target.plugin import Plugin, export

if TYPE_CHECKING:
    from collections.abc import Iterator


class LocalePlugin(Plugin):
    """Default locale plugin."""

    def check_compatible(self) -> None:
        pass

    @export(property=True)
    def timezone(self) -> str | None:
        """Get the timezone of the system."""
        return None

    @export(property=True)
    def language(self) -> list[str]:
        """Get the configured locale(s) of the system."""
        return []

    @export(record=EmptyRecord)
    def keyboard(self) -> Iterator[EmptyRecord]:
        """Get the keyboard layout(s) of the system."""
        yield from []



from __future__ import annotations

import json
import re
import urllib.parse
from functools import lru_cache
from typing import TYPE_CHECKING

from dissect.target.exceptions import UnsupportedPluginError
from dissect.target.helpers.record import DynamicDescriptor, TargetRecordDescriptor
from dissect.target.plugin import Plugin, arg, export

if TYPE_CHECKING:
    from collections.abc import Iterator

    from flow.record import Record

    from dissect.target.target import Target

VELOCIRAPTOR_RESULTS = "/$velociraptor_results$"
ISO_8601_PATTERN = r"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(\.\d+)?(Z|[+-]\d{2}:\d{2})?"


class VelociraptorRecordBuilder:
    def __init__(self, artifact_name: str, extract_nested: bool):
        self._create_event_descriptor = lru_cache(4096)(self._create_event_descriptor)
        self.record_name = f"velociraptor/{artifact_name}"
        self.extract_nested = extract_nested

    def build(self, object: dict, target: Target) -> TargetRecordDescriptor:
        """Builds a Velociraptor record."""
        record_values = {}
        record_fields = []

        record_values["_target"] = target

        for key, value in object.items():
            # Reserved by flow.record
            if key.startswith("_"):
                continue

            key = key.lower().replace("(", "_").replace(")", "_")

            if re.match(ISO_8601_PATTERN, str(value)):
                record_type = "datetime"
            elif isinstance(value, list):
                record_type = "string[]"
            elif isinstance(value, int):
                record_type = "varint"
            elif key == "hash":
                record_type = "digest"
                value = (value.get("MD5"), value.get("SHA1"), value.get("SHA256"))
            elif isinstance(value, str):
                record_type = "string"
            elif isinstance(value, dict):
                if self.extract_nested:
                    record_type = "record"
                    value = self.build(value, target)
                else:
                    # Skip nested objects that contain additional metadata
                    continue
            else:
                record_type = "dynamic"

            record_fields.append((record_type, key))
            record_values[key] = value

        # tuple conversion here is needed for lru_cache
        desc = self._create_event_descriptor(tuple(record_fields))
        return desc(**record_values)

    def _create_event_descriptor(self, record_fields: list[tuple[str, str]]) -> TargetRecordDescriptor:
        return TargetRecordDescriptor(self.record_name, record_fields)


class VelociraptorPlugin(Plugin):
    """Returns records from Velociraptor artifacts."""

    __namespace__ = "velociraptor"

    def __init__(self, target: Target):
        super().__init__(target)
        self.results_dir = target.fs.path(VELOCIRAPTOR_RESULTS)

    def check_compatible(self) -> None:
        if not self.results_dir.exists():
            raise UnsupportedPluginError("No Velociraptor artifacts found")

    @export(record=DynamicDescriptor(["datetime"]))
    @arg(
        "--extract-nested",
        action="store_true",
        help="extracts JSON objects from the artifacts",
    )
    def results(
        self,
        extract_nested: bool = False,
    ) -> Iterator[Record]:
        """Return Rapid7 Velociraptor artifacts.

        By default JSON objects are not extracted from the artifacts,
        this can be done with the argument ``--extract-nested``.

        References:
            - https://docs.velociraptor.app/docs/vql/artifacts/
        """
        for artifact in self.results_dir.glob("*.json"):
            # "Windows.KapeFiles.Targets%2FAll\ File\ Metadata.json" becomes "windows_kapefiles_targets"
            artifact_name = (
                urllib.parse.unquote(artifact.name.removesuffix(".json")).split("/")[0].lower().replace(".", "_")
            )
            record_builder = VelociraptorRecordBuilder(artifact_name, extract_nested=extract_nested)

            for line in artifact.open("rt"):
                if not (line := line.strip()):
                    continue

                try:
                    object = json.loads(line)
                    yield record_builder.build(object, self.target)
                except json.decoder.JSONDecodeError:
                    self.target.log.warning(
                        "Could not decode Velociraptor JSON log line in file %s: %s",
                        artifact,
                        line,
                    )
                    continue
